# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Defines a Celestial Body, which acts as the central repository of information for
the planets or similar Central Bodies.

"""
from astropy import units as u
from numpy import inf

from satmad.core.central_body import CelestialBody, CelestialBodyEllipsoid

# **************** GM values ****************

GM_earth = 3.986004418e14 * u.m ** 3 / u.s ** 2
"""Geocentric gravitational constant
(TCG-compatible value).

(See IERS Technical Note No. 36 (2010) Table 1.1 [TCF1]_)"""

GM_sun = 1.32712442099e20 * u.m ** 3 / u.s ** 2
"""Heliocentric gravitational constant
(TCB-compatible value, computed from the TDB-compatible value).

(See IERS Technical Note No. 36 (2010) Table 1.1  [TCF1]_)"""

GM_moon = 4.902802711497899e12 * u.m ** 3 / u.s ** 2
r"""Lunar  gravitational constant.

The value is derived from "Moon-Earth mass ratio" times "Geocentric
gravitational constant" (:math:`0.0123000371 \times 398600.64418`).

(See IERS Technical Note No. 36 (2010) Table 1.1  [TCF1]_)."""

# **************** Ellipsoid Definitions ****************

EARTH_ELLIPSOID_IERS2003 = CelestialBodyEllipsoid(
    "Earth Ellipsoid IERS 2003",
    6378136.6 * u.m,
    298.25642 * u.dimensionless_unscaled,
    mu=3.986004418 * u.m ** 3 / u.s ** 2,
    j2=1.0826359e-3 * u.dimensionless_unscaled,
)
"""Earth Ellipsoid defined in IERS 2010 Numerical Standards.

(See IERS Technical Note No. 36 (2010) Table 1.1  [TCF1]_)."""

EARTH_ELLIPSOID_GRS80 = CelestialBodyEllipsoid(
    "Earth Ellipsoid GRS80",
    6378137 * u.m,
    298.257222101 * u.dimensionless_unscaled,
    mu=3.986005e14 * u.m ** 3 / u.s ** 2,
    j2=1.08263e-3 * u.dimensionless_unscaled,
    om=7.292115e-5 * u.rad / u.s,
)
"""Earth Ellipsoid defined in Geodetic Reference System GRS80.

(See IERS Technical Note No. 36 (2010) Table 1.2  [TCF1]_)."""

EARTH_ELLIPSOID_WGS84 = CelestialBodyEllipsoid(
    "Earth Ellipsoid WGS84", 6378137 * u.m, 298.257223563 * u.dimensionless_unscaled
)
"""Earth Ellipsoid defined in World Geodetic System WGS84.

(See `WGS84 definition <https://earth-info.nga.mil/GandG/publications/tr8350.2/tr8350_2.html>`_)."""

MOON_ELLIPSOID_IAUWG2015 = CelestialBodyEllipsoid(
    "Moon Ellipsoid IAU WG 2015", 1737400 * u.m, inf * u.dimensionless_unscaled
)
"""Moon Ellipsoid - flattening is zero as the reference value means a perfect sphere.

(See Report of the IAU Working Group on Cartographic Coordinates and Rotational Elements
(2015) Table 5  [TCF2]_)."""

# **************** Celestial Body Definitions ****************

EARTH = CelestialBody(
    "Earth",
    "Default Earth Model. ",
    GM_earth.to(u.km ** 3 / u.s ** 2),
    inert_coord="gcrs",
    body_fixed_coord="itrs",
    ellipsoid=EARTH_ELLIPSOID_GRS80,
)
"""Default Earth Model."""

SUN = CelestialBody(
    "Sun", "Default Sun Model. ", GM_sun.to(u.km ** 3 / u.s ** 2), inert_coord="hcrs"
)
"""Default Sun Model."""

MOON = CelestialBody(
    "Moon",
    "Default Moon Model. ",
    GM_moon.to(u.km ** 3 / u.s ** 2),
    ellipsoid=MOON_ELLIPSOID_IAUWG2015,
    inert_coord="mooncrs",
)
"""Default Moon Model."""



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Base module for numerical propagators.

"""
from enum import Enum

import numpy as np
from astropy import units as u
from astropy.coordinates import (
    GCRS,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from scipy.integrate import solve_ivp

from satmad.coordinates.trajectory import Trajectory
from satmad.core.celestial_bodies import EARTH
from satmad.propagation.base_propagator import AbstractPropagator
from satmad.propagation.force_models import two_body_accel


class ODESolverType(Enum):
    """
    Built-in solver types for :meth:`~scipy.integrate.solve_ivp`.
    """

    RK45 = "RK45"
    RK23 = "RK23"
    DOP853 = "DOP853"
    RADAU = "Radau"
    BDF = "BDF"
    LSODA = "LSODA"


class NumericalPropagator(AbstractPropagator):
    """
    Numerical orbit propagator based on Scipy ODE Solvers.

    Analysis shows that `DOP853` is currently the best (and default) choice for
    accuracy and runtime performance.

    Parameters
    ----------
    name : str
        Name or identifier for the propagator
    stepsize : Quantity or `~astropy.time.TimeDelta`
        Output stepsize for the propagator
    solver_type : ODESolverType
        Solver type to be used
    rtol : float
        relative tolerance value
    atol : float
        absolute tolerance value
    central_body : CelestialBody
        Centre Celestial Body for the propagator
    """

    def __init__(
        self,
        stepsize=60 * u.s,
        solver_type=ODESolverType.DOP853,
        rtol=1e-12,
        atol=1e-14,
        name="",
        central_body=EARTH,
    ):
        super().__init__(stepsize, name, central_body)

        self._solver_type = solver_type

        self.rtol = rtol
        self.atol = atol

        if not name:
            self._name = f"{solver_type.value} (rtol: {rtol}, atol: {atol} )"

    def gen_trajectory(self, init_coords, interval):
        """
        Generates the trajectory (time and coordinate information) for the given
        interval with the internal output stepsize.

        If the initial coordinate is in a different frame than the propagation frame,
        it is automatically converted to the proper frame for propagation.

        Parameters
        ----------
        init_coords : SkyCoord
            Initial coordinates (the first value is used)
        interval : TimeInterval
            Time interval for which the ephemeris will be generated
        Returns
        -------
        Trajectory
            The output trajectory
        """
        # generate the output timelist
        time_list = self._generate_time_list(interval)

        # Propagate the orbit
        if init_coords.isscalar:
            # Single SkyCoord element available
            init_coords_converted = init_coords
        else:
            # Multiple SkyCoord elements available, use first
            init_coords_converted = init_coords[0]

        # Check for the initial coordinate system, convert if necessary
        if self.central_body.inert_coord != init_coords_converted.frame.name:
            init_coords_converted = init_coords_converted.transform_to(
                self.central_body.inert_coord
            )

        # Run propagation
        coords_list = self._run_propagation(init_coords_converted, time_list)

        # generate the trajectory
        return Trajectory(coords_list)

    def _run_propagation(self, init_coords, time_list):
        """Solves the ODE or runs the actual propagation.

        Parameters
        ----------
        init_coords : SkyCoord
            Initial coordinates (the first value is used)
        time_list : Time
            Time list for outputs
        Returns
        -------
        SkyCoord
            The output coordinates corresponding to `time_list`
        """

        # generate the time list since epoch (in seconds)
        t_list_epoch = (time_list - time_list[0]).to_value(u.s)

        # concat the init vector
        rv_init = [
            *init_coords.cartesian.without_differentials().xyz.to_value(u.km),
            *init_coords.velocity.d_xyz.to_value(u.km / u.s),
        ]

        solver = solve_ivp(
            self._ode_diff_eqns,
            (t_list_epoch[0], t_list_epoch[-1]),
            rv_init,
            method=self._solver_type.value,
            dense_output=True,
            rtol=self.rtol,
            atol=self.atol,
        )

        r_list = np.empty((3, len(t_list_epoch)))
        v_list = np.empty((3, len(t_list_epoch)))
        # Fill the raw output data
        i = 0
        for t in t_list_epoch:
            rv = solver.sol(t)
            r_list[:, i] = rv[:3]
            v_list[:, i] = rv[3:]
            i = i + 1

        # Load the time, pos, vel info into astropy objects (shallow copied)
        rv_list = CartesianRepresentation(r_list, unit=u.km).with_differentials(
            CartesianDifferential(v_list, unit=u.km / u.s)
        )

        coords_list = SkyCoord(
            rv_list,
            obstime=time_list,
            frame=GCRS,
            representation_type="cartesian",
            differential_type="cartesian",
            copy=False,
        )
        return coords_list

    def _ode_diff_eqns(self, t, rv):
        """
        Defines the Ordinary Differential Equations that govern the motion of the
        satellite. In practice, this method computes the instantaneous accelerations
        acting on the satellite.

        Note that the input `rv` store the position and velocity vectors in the
        inertial frame of the central body. For the Earth this is GCRS.

        This method is used by the scipy `solve_ivp()` solver to evaluate the differential
        equations at a given time.

        Parameters
        ----------
        t : Time
            Time where the accelerations are defined
        rv : ndarray
            array of 6 values, the first three is the position vector (km) and
            the second three is the velocity vector (km/s)

        Returns
        -------
        ndarray
            The array with velocity and acceleration information.
        """

        r = rv[:3]
        v = rv[3:]

        mu = self.central_body.mu

        # two-body acceleration
        a = two_body_accel(r, mu.to_value(_km3s2))

        # concat the two lists using PEP-448 unpack
        return [*v, *a]


_km3s2 = u.km ** 3 / u.s ** 2



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Tests associated with the numerical propagators.

"""
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import (
    GCRS,
    ITRS,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from astropy.time import Time
from pytest import approx

from satmad.coordinates.frames import MoonCRS
from satmad.coordinates.tests.test_baseframe_transform import pos_err, vel_err
from satmad.core.central_body import CelestialBody, CelestialBodyEllipsoid
from satmad.propagation.numerical_propagators import ODESolverType
from satmad.propagation.tests.num_prop_analysis_engine import (
    energy_along_trajectory,
    propagation_engine,
)


@pytest.fixture
def rv_init_moon_crs():
    """Initialises a Moon low-orbit satellite."""
    time = Time("2020-01-01T11:00:00.000", scale="utc")

    # GMAT test case
    v_moon_crs = CartesianDifferential([1, -1, 0.6], unit=u.km / u.s)
    r_moon_crs = CartesianRepresentation([1000, 1000, 2000], unit=u.km)
    rv_init = SkyCoord(
        r_moon_crs.with_differentials(v_moon_crs),
        obstime=time,
        frame=MoonCRS,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    return rv_init


_MOONGMAT = CelestialBody(
    "Moon",
    "Default Moon Model. ",
    4.90280105600000e12 * u.m ** 3 / u.s ** 2,
    ellipsoid=CelestialBodyEllipsoid(
        "Moon Ellipsoid as defined by GMAT",
        1738.2 * u.km,
        np.inf * u.dimensionless_unscaled,
    ),
    inert_coord="mooncrs",
)
"""This is the Moon as defined by GMAT"""


def test_num_propagator_moon(rv_init_moon_crs):
    """This tests the final position and velocity against GMAT using a pure Keplerian
    5 day propagation.

    GMAT Propagator is Dormand-Prince 853, with an Accuracy of 1e-13."""

    allowable_pos_diff = 13 * u.mm
    allowable_vel_diff = 0.0055 * u.mm / u.s

    # Set up propagation config
    stepsize = 60 * u.s
    solver_type = ODESolverType.DOP853
    rtol = 1e-13
    atol = 1e-15
    init_time_offset = 0.0 * u.day
    duration = 10.00 * u.day

    # run propagation and get trajectory
    trajectory = propagation_engine(
        rv_init_moon_crs,
        stepsize,
        solver_type,
        init_time_offset,
        duration,
        rtol,
        atol,
        central_body=_MOONGMAT,
    )
    # no interpolation, just take final element
    rv_fin = trajectory.coord_list[-1]

    # GMAT test case - final point
    time = Time("2020-01-11T11:00:00.000", scale="utc")

    v_moon_crs_gmat = CartesianDifferential(
        [6.159801708156309e-01, -1.128605958878207e00, 1.075005007078085e-02],
        unit=u.km / u.s,
    )
    r_moon_crs_gmat = CartesianRepresentation(
        [1.870952777669765e03, -1.811234513779054e02, 2.305452195816935e03],
        unit=u.km,
    )
    rv_fin_gmat = SkyCoord(
        r_moon_crs_gmat.with_differentials(v_moon_crs_gmat),
        obstime=time,
        frame=MoonCRS,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    r_diff = pos_err(rv_fin, rv_fin_gmat)
    v_diff = vel_err(rv_fin, rv_fin_gmat)

    print(f"r diff      :  {r_diff}")
    print(f"v diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


@pytest.fixture
def rv_init_leo_gcrs():
    """Initialises an Earth LEO satellite."""
    time = Time("2004-04-06T07:51:28.386009", scale="utc")

    # Vallado IAU 2000 - Table 3-6
    v_gcrs = CartesianDifferential(
        [-4.7432201610, 0.7905364950, 5.5337557240], unit=u.km / u.s
    )
    r_gcrs = CartesianRepresentation(
        [5102.50895290, 6123.01139910, 6378.13693380], unit=u.km
    )
    rv_init = SkyCoord(
        r_gcrs.with_differentials(v_gcrs),
        obstime=time,
        frame=GCRS,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    return rv_init


def test_num_propagator(rv_init_leo_gcrs):
    """Tests the basic numerical propagation accuracy in terms of energy."""

    # Set up propagation config
    # Start with ITRS to test coord transform on propagator entry
    rv_init = rv_init_leo_gcrs.transform_to(ITRS)
    stepsize = 60 * u.s
    solver_type = ODESolverType.DOP853
    rtol = 1e-13
    atol = 1e-15
    init_time_offset = 0.5 * u.day
    duration = 10.001 * u.day

    # run propagation and get trajectory
    trajectory = propagation_engine(
        rv_init, stepsize, solver_type, init_time_offset, duration, rtol, atol
    )
    # print(trajectory.coord_list[-1])

    # Compute the energy along the trajectory
    time_list, energy_list, energy_diff_list = energy_along_trajectory(
        trajectory.coord_list
    )

    abs_energy_diff_list = abs(np.array(energy_diff_list))

    # Max energy diff (abs value) - should be low
    energy_diff_max = abs_energy_diff_list.max()
    # Mean energy diff - should be zero to have zero mean energy
    energy_diff_mean = np.array(energy_diff_list).mean()

    # initial energy diff and final energy diff to assess energy leak
    # First and last 100 elements taken and they should also ideally be zero
    init_energy_diff_mean = np.array(energy_diff_list)[:100].mean()
    final_energy_diff_mean = np.array(energy_diff_list)[-100:].mean()

    # print(f"max energy diff: {energy_diff_max}")
    # print(f"mean energy    : {energy_diff_mean}")
    #
    # print(f"init mean energy diff : {init_energy_diff_mean}")
    # print(f"final mean energy diff: {final_energy_diff_mean}")

    assert energy_diff_max < 4.50e-11
    assert energy_diff_mean < 2.95e-12

    assert init_energy_diff_mean < 1.25e-12
    assert final_energy_diff_mean < 5.20e-12



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Trajectory Class tests.
"""


import numpy as np
from astropy import units as u
from astropy.coordinates import (
    TEME,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from astropy.time import Time, TimeDelta
from matplotlib import pyplot as plt
from sgp4.api import Satrec

from satmad.coordinates.trajectory import Trajectory
from satmad.propagation.tle import TLE


def generate_timelist(init_time, duration, steps):
    """generates the time lists for propagation."""
    # ****** Generate the discrete time instants through the propagation duration ******
    dt = duration / steps

    # Generate time list
    dt_list = dt * np.arange(0, steps, 1)
    time_list = init_time + dt_list

    return dt, time_list


def propagation_engine(line1, line2, time_list):
    """Tests the interpolation accuracy for a given TLE."""

    # Init satellite object from the TLE
    sat = Satrec.twoline2rv(line1, line2)

    # ****** Generate the pos, vel vectors for each time instant ******

    # Run the propagation and init pos and vel vectors in TEME
    e, r_list, v_list = sat.sgp4_array(time_list.jd1, time_list.jd2)

    # Load the time, pos, vel info into astropy objects (shallow copied)
    vel_list = CartesianDifferential(v_list, unit=u.km / u.s, xyz_axis=1)
    pos_list = CartesianRepresentation(
        r_list, unit=u.km, xyz_axis=1
    ).with_differentials(vel_list)

    # trajectory in astropy
    traj_astropy = SkyCoord(
        pos_list,
        obstime=time_list,
        frame=TEME.name,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    # Init trajectory in Trajectory object
    trajectory = Trajectory(traj_astropy)

    return {"traj": trajectory, "sat": sat}


def check_errors(init_time, dt, sat, trajectory, generate_plots=False):
    """Checks the errors for the given trajectory."""
    # **** Check the results ****
    step_offset = 0
    test_stepsize = 1.0 * u.s
    test_end = 0.01 * u.day

    # print(f"Test duration: {(test_end - step_offset * dt).to(u.day)}")
    # print(f"Step size: {test_stepsize}")

    dt_list_hires = TimeDelta(
        np.arange(
            step_offset * dt.to_value(u.s),
            test_end.to_value(u.s),
            test_stepsize.to_value(u.s),
        ),
        format="sec",
    )
    time_list_hires = init_time + dt_list_hires

    # Generate the high-res test trajectory
    e_test_list, r_test_list, v_test_list = sat.sgp4_array(
        time_list_hires.jd1, time_list_hires.jd2
    )

    r_test_list = CartesianRepresentation(r_test_list.transpose(), unit=u.km)

    # Generate the error list
    r_err_list = (
        trajectory(time_list_hires).cartesian.without_differentials() - r_test_list
    )

    # Error stats
    max_err = r_err_list.norm().max()  # max expected error
    max_nominal_err = r_err_list[-100:].norm().max()  # max expected error nominal

    if generate_plots:
        print(f"max error at the beginning  : {max_err.to(u.mm)}")
        print(f"max error after settle down : {max_nominal_err.to(u.mm)}")

        dt_list_hires_plot = dt_list_hires.to_value(u.day)

        # Plot the error per axis
        fr_ticker = np.arange(
            dt_list_hires_plot[0], dt_list_hires_plot[-1], dt.to_value(u.day)
        )
        y_ticker = np.zeros(len(fr_ticker))
        plt.figure()
        plt.plot(
            dt_list_hires_plot,
            np.asarray(r_err_list.get_xyz(xyz_axis=1)),
            fr_ticker,
            y_ticker,
            "x",
        )
        plt.legend(["x", "y", "z", "stepsize"])
        plt.title(f"Interpolation Error\n({trajectory.interpolator_name})")
        plt.xlabel("Time [days]")
        plt.ylabel("Error [km]")
        # plt.yscale("log")  # Uncomment to get y axis in log scale
        plt.show()

    return max_err, max_nominal_err


def test_trajectory_export():
    """Tests the interpolation accuracy for ISS."""

    # Init TLE
    line1 = "1 25544U 98067A   19343.69339541  .00001764  00000-0  38792-4 0  9991"
    line2 = "2 25544  51.6439 211.2001 0007417  17.6667  85.6398 15.50103472202482"

    # test init time
    init_time = Time(2458826.3, format="jd", scale="utc")
    init_time.format = "isot"

    duration = TimeDelta(3.0, format="jd")
    steps = 4000  # number of steps within the propagation duration

    dt, time_list = generate_timelist(init_time, duration, steps)
    prop_output = propagation_engine(line1, line2, time_list)

    ts = prop_output["traj"].to_timeseries()

    pvt = prop_output["traj"].coord_list[10]

    # print(ts)
    # print(ts["time", "v_X"])

    assert (ts.time[10] - pvt.obstime) < 17 * u.us
    assert (ts[10]["r_X"] - pvt.cartesian.x) < 1 * u.um


def test_interpolation_err_iss():
    """Tests the interpolation accuracy for ISS."""

    # Init TLE
    line1 = "1 25544U 98067A   19343.69339541  .00001764  00000-0  38792-4 0  9991"
    line2 = "2 25544  51.6439 211.2001 0007417  17.6667  85.6398 15.50103472202482"

    # test init time
    init_time = Time(2458826.3, format="jd", scale="utc")

    duration = TimeDelta(3.0, format="jd")
    steps = 4000  # number of steps within the propagation duration

    dt, time_list = generate_timelist(init_time, duration, steps)
    prop_output = propagation_engine(line1, line2, time_list)
    max_begin_err, max_nominal_err = check_errors(
        init_time, dt, prop_output["sat"], prop_output["traj"], False
    )

    assert max_begin_err < 16 * u.mm
    assert max_nominal_err < 0.08 * u.mm


def test_interpolation_err_geo():
    """Tests the interpolation accuracy for GEO."""

    # Init TLE
    line1 = "1 99999U 12345A   20162.50918981  .00000000  00000-0  00000-0 0 00005"
    line2 = "2 99999 000.0000 124.6202 0000000 000.0000 000.0000 01.00273791000004"

    # test init time
    init_time = Time("2020:162:09:00:00", format="yday", scale="utc")

    duration = TimeDelta(3.0, format="jd")
    steps = 4000  # number of steps within the propagation duration

    dt, time_list = generate_timelist(init_time, duration, steps)
    prop_output = propagation_engine(line1, line2, time_list)
    max_err, max_nominal_err = check_errors(
        init_time, dt, prop_output["sat"], prop_output["traj"], False
    )

    assert max_err < 0.0005 * u.mm
    assert max_nominal_err < 0.0003 * u.mm


def test_interpolation_err_heo():
    """Tests the interpolation accuracy for Highly Elliptic Orbits."""

    # Init TLE
    line1 = "1 99999U 12345A   20162.50918981  .00000000  00000-0  00000-0 0 00005"
    line2 = "2 99999 000.0000 124.6202 0000000 000.0000 000.0000 01.00273791000004"

    # test init time
    init_time = Time("2020:162:00:00:00", format="yday", scale="utc")

    tle = TLE.from_tle(line1, line2)

    tle.epoch = init_time
    tle.mean_anomaly = 0
    tle.eccentricity = 0.4

    heo_line1, heo_line2 = tle.export_tle()

    # print(tle)

    duration = TimeDelta(3.0, format="jd")
    steps = 4000  # number of steps within the propagation duration

    dt, time_list = generate_timelist(init_time, duration, steps)
    prop_output = propagation_engine(heo_line1, heo_line2, time_list)
    max_err, max_nominal_err = check_errors(
        init_time, dt, prop_output["sat"], prop_output["traj"], False
    )

    assert max_err < 0.040 * u.mm
    assert max_nominal_err < 0.00030 * u.mm



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------

project = "satmad"
copyright = "2021, Egemen Imre"
author = "Egemen Imre"

# Version Info
# ------------
# The short X.Y version.
version = "0.1"
# The full version, including alpha/beta/rc tags.
release = "0.1"

# -- General configuration ---------------------------------------------------
# By default, highlight as Python 3.
highlight_language = "python3"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # auto api generation
    "sphinx.ext.napoleon",  # numpy support
    "sphinx.ext.mathjax",  # LaTex style math
    "sphinx.ext.graphviz",  # Dependency diagrams
    "sphinx.ext.intersphinx",  # Link mapping to external projects
    # "sphinx.ext.doctest",  # Doctest
    "nbsphinx",  # Jupyter notebook support
]
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# Intersphinx configuration
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("https://matplotlib.org", None),
}

if os.environ.get("READTHEDOCS") == "True":
    nbsphinx_execute = "never"
else:
    nbsphinx_execute = "always"

# Controls when a cell will time out (defaults to 30; use -1 for no timeout):
nbsphinx_timeout = 1000

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Methods to analyse and measure numerical propagation performance.

"""
from time import perf_counter

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.visualization import quantity_support, time_support
from matplotlib import pyplot as plt

from satmad.coordinates.trajectory import Trajectory
from satmad.core.celestial_bodies import EARTH, GM_earth
from satmad.propagation.force_models import two_body_energy
from satmad.propagation.numerical_propagators import NumericalPropagator
from satmad.propagation.sgp4_propagator import SGP4Propagator
from satmad.propagation.tle import TLE
from satmad.utils.timeinterval import TimeInterval


def tle_heo():
    """Generates the TLE with HEO test setup (e=0.70)."""

    name = "ATLAS 2A CENTAUR R/B"
    line1 = "1 23840U 96020B   18198.38669861 -.00000081  00000-0  00000+0 0  9997"
    line2 = "2 23840  21.5075 284.9295 6975939  94.1963 356.6140  2.24211004182606"

    return TLE.from_tle(line1, line2, name)


def tle_leo():
    """Generates the TLE with near-circular LEO test setup."""

    name = "RASAT"
    line1 = "1 37791U 11044D   18198.20691930 -.00000011  00000-0  70120-5 0  9992"
    line2 = "2 37791  98.1275 290.4108 0021116 321.0704  38.8990 14.64672859369594"

    return TLE.from_tle(line1, line2, name)


def tle_geo():
    """Generates the TLE with near-circular LEO test setup."""

    name = "TURKSAT 4B"
    line1 = "1 40984U 15060A   18198.04228921  .00000095  00000-0  00000+0 0  9995"
    line2 = "2 40984   0.0142 325.6420 0001708 151.3463 243.0113  1.00269765 10118"

    return TLE.from_tle(line1, line2, name)


def init_rv(tle, epoch):
    """Initialises the position and velocity vectors at the requested epoch."""
    # init SGP4
    return SGP4Propagator().propagate(tle, epoch)


def propagation_engine(
    rv_init,
    stepsize,
    solver_type,
    init_time_offset,
    duration,
    rtol,
    atol,
    central_body=EARTH,
):
    """Initialises and runs the propagation engine to yield the resulting trajectory."""

    # init propagator
    prop = NumericalPropagator(
        stepsize,
        solver_type=solver_type,
        rtol=rtol,
        atol=atol,
        name="",
        central_body=central_body,
    )

    prop_start = rv_init.obstime + init_time_offset
    prop_duration = duration

    trajectory = prop.gen_trajectory(rv_init, TimeInterval(prop_start, prop_duration))

    return trajectory


def energy_along_trajectory(
    coord_list: SkyCoord, mu=GM_earth.to_value(u.km ** 3 / u.s ** 2)
):
    """Computes the absolute and relative (to initial) energy along the trajectory."""

    time_list = coord_list.obstime - coord_list[0].obstime

    init_coords = coord_list[0]
    init_energy = two_body_energy(
        init_coords.cartesian.without_differentials().xyz.to_value(u.km),
        init_coords.velocity.d_xyz.to_value(u.km / u.s),
        mu,
    )

    energy_list = []
    energy_diff_list = []
    for coords in coord_list:
        r = coords.cartesian.without_differentials().xyz.to_value(u.km)
        v = coords.velocity.d_xyz.to_value(u.km / u.s)
        energy = two_body_energy(r, v, mu)
        energy_list.append(energy)
        energy_diff_list.append(energy - init_energy)

    return time_list, energy_list, energy_diff_list


def plot_energy_diff(cases, title, logscale_plot=True):
    """Plots the energy difference."""
    quantity_support()
    time_support()

    fig, ax = plt.subplots()

    ax.grid()
    # ax.xaxis.set_major_formatter(FormatStrFormatter("% 3.3f"))
    # ax.yaxis.set_major_formatter(FormatStrFormatter("% 3.3f"))
    plt.title(f"Specific Energy Evolution wrt. Initial State\n({title})")
    ax.set_xlabel("Time (days)")
    ax.set_ylabel("Specific Energy Diff (m^2/s^2)")

    if logscale_plot:
        ax.set_yscale("log")

    fig.set_size_inches(8, 6)

    for case in cases:
        if logscale_plot:
            # Make sure everything is positive for plotting
            ax.plot(
                cases[case]["time_list"].to(u.day),
                np.abs(cases[case]["energy_diff_list"]),
                label=case,
            )
        else:
            ax.plot(
                cases[case]["time_list"].to(u.day),
                cases[case]["energy_diff_list"],
                label=case,
            )

    plt.legend()
    plt.show()


def plot_runtimes(cases, title, logscale_plot=True):
    """Plots the runtimes."""
    quantity_support()
    time_support()

    fig, ax = plt.subplots()

    # ax.grid()
    # ax.xaxis.set_major_formatter(FormatStrFormatter("% 3.3f"))
    # ax.yaxis.set_major_formatter(FormatStrFormatter("% 3.3f"))
    plt.title(f"Runtime Analysis\n({title})")
    ax.set_ylabel("Runtime (s)")

    if logscale_plot:
        ax.set_yscale("log")

    fig.set_size_inches(8, 6)

    # generate data lists
    names = [case for case in cases]
    runtimes = [cases[case]["runtime"] for case in cases]

    # generate the plots
    if logscale_plot:
        # Make sure everything is positive for plotting
        ax.bar(names, [np.abs(runtime) for runtime in runtimes])
    else:
        ax.bar(names, runtimes)

    # Annotate the plot
    for x, y in zip(names, runtimes):
        label = "{:.2f}".format(y)
        plt.annotate(
            label,  # this is the text
            (x, y),  # this is the point to label
            textcoords="offset points",  # how to position the text
            xytext=(0, 5),  # distance from text to points (x,y)
            ha="center",
        )  # horizontal alignment can be left, right or center

    # ax.legend([case for case in cases])
    plt.show()


def analyse_energy_along_trajectory(
    cases, name, rv_init, stepsize, solver_type, init_time_offset, duration, rtol, atol
):
    """Analyses the energy evolution along the trajectory."""

    print(f"**** Case: {name} ****")

    tic = perf_counter()

    # run propagation and get trajectory
    trajectory = propagation_engine(
        rv_init, stepsize, solver_type, init_time_offset, duration, rtol, atol
    )

    toc = perf_counter()

    runtime = (toc - tic) * u.s

    # Compute the energy along the trajectory
    time_list, energy_list, energy_diff_list = energy_along_trajectory(
        trajectory.coord_list
    )
    cases[name] = {
        "traj": trajectory,
        "time_list": time_list,
        "energy_list": energy_list,
        "energy_diff_list": energy_diff_list,
        "runtime": runtime,
    }

    print(f"Runtime: {runtime:0.4f} seconds")

    # Print summary results

    abs_energy_diff_list = abs(np.array(energy_diff_list))

    # Max energy diff (abs value) - should be low
    energy_diff_max = abs_energy_diff_list.max()
    # Mean energy diff - should be zero to have zero mean energy
    energy_diff_mean = np.array(energy_diff_list).mean()

    # initial energy diff and final energy diff to assess energy leak
    # First and last 100 elements taken and they should also ideally be zero
    init_energy_diff_mean = np.array(energy_diff_list)[:100].mean()
    final_energy_diff_mean = np.array(energy_diff_list)[-100:].mean()

    print(f"max energy diff: {energy_diff_max}")
    print(f"mean energy    : {energy_diff_mean}")

    print(f"init mean energy diff : {init_energy_diff_mean}")
    print(f"final mean energy diff: {final_energy_diff_mean}")


def pos_along_trajectory(coord_list: SkyCoord, truth_traj: Trajectory):
    """Computes the absolute and relative (to "truth") position along the trajectory."""

    time_list = coord_list.obstime - coord_list[0].obstime

    pos_diff_list = []
    for coords in coord_list:
        r = coords.cartesian.without_differentials()
        # v = coords.velocity.d_xyz.to_value(u.km / u.s)

        r_truth = truth_traj(coords.obstime).cartesian.without_differentials()

        pos_diff_list.append((r - r_truth).norm().to_value(u.m))

    return time_list, pos_diff_list


def analyse_pos_along_trajectory(
    truth_traj,
    cases,
    name,
    rv_init,
    stepsize,
    solver_type,
    init_time_offset,
    duration,
    rtol,
    atol,
):
    """Analyses the position variation along the trajectory."""

    print(f"**** Case: {name} ****")

    tic = perf_counter()

    # run propagation and get trajectory
    trajectory = propagation_engine(
        rv_init, stepsize, solver_type, init_time_offset, duration, rtol, atol
    )

    toc = perf_counter()

    runtime = (toc - tic) * u.s

    # Compute the energy along the trajectory
    time_list, pos_diff_list = pos_along_trajectory(trajectory.coord_list, truth_traj)
    cases[name] = {
        "traj": trajectory,
        "time_list": time_list,
        "pos_diff_list": pos_diff_list,
        "runtime": runtime,
    }

    print(f"Runtime: {runtime:0.4f} seconds")

    # Print summary results

    abs_pos_diff_list = abs(np.array(pos_diff_list))

    # Max pos diff (abs value) - should be low
    pos_diff_max = abs_pos_diff_list.max()
    # # Mean pos diff - should be zero to have zero mean energy
    # pos_diff_mean = np.array(pos_diff_list).mean()
    #
    # # initial pos diff and final pos diff to assess energy leak
    # # First and last 100 elements taken and they should also ideally be zero
    # init_pos_diff_mean = np.array(pos_diff_list)[:100].mean()
    # final_pos_diff_mean = np.array(pos_diff_list)[-100:].mean()

    print(f"max pos diff: {pos_diff_max} m")
    # print(f"mean pos diff   : {pos_diff_mean}")
    #
    # print(f"init mean energy diff : {init_pos_diff_mean}")
    # print(f"final mean energy diff: {final_pos_diff_mean}")


def plot_pos_diff(cases, title, logscale_plot=True):
    """Plots the pos difference."""
    quantity_support()
    time_support()

    fig, ax = plt.subplots()

    ax.grid()
    # ax.xaxis.set_major_formatter(FormatStrFormatter("% 3.3f"))
    # ax.yaxis.set_major_formatter(FormatStrFormatter("% 3.3f"))
    plt.title(f"Position Difference Variation\n({title})")
    ax.set_xlabel("Time (days)")
    ax.set_ylabel("Pos Diff (m)")

    if logscale_plot:
        ax.set_yscale("log")

    fig.set_size_inches(8, 6)

    for case in cases:
        if logscale_plot:
            # Make sure everything is positive for plotting
            ax.plot(
                cases[case]["time_list"].to(u.day),
                np.abs(cases[case]["pos_diff_list"]),
                label=case,
            )
        else:
            ax.plot(
                cases[case]["time_list"].to(u.day),
                cases[case]["pos_diff_list"],
                label=case,
            )

    ax.legend()
    plt.show()



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
SGP4 propagator tests.

"""

import pytest
from astropy import units as u
from astropy.coordinates import (
    GCRS,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)

from satmad.coordinates.frames import J2000
from satmad.propagation.sgp4_propagator import (
    SGP4GeneralError,
    SGP4Propagator,
    SGP4SatDecayedError,
)
from satmad.propagation.tle import TLE
from satmad.utils.timeinterval import TimeInterval


@pytest.fixture
def init_tle_decay():
    """Generates the TLE with decay test setup."""

    name = "MICROSAT-R DEB"
    line1 = "1 44160U 19006AX  20178.55672017  .00609591  12600-3  17697-2 0  9996"
    line2 = "2 44160  95.2356 283.8798 0142512 327.7074  31.5529 15.78083811 64981"

    return TLE.from_tle(line1, line2, name)


@pytest.fixture
def init_tle_vgd():
    """Generates the TLE with LEO test setup."""

    name = "VANGUARD 1"
    line1 = "1 00005U 58002B   00179.78495062  .00000023  00000-0  28098-4 0  4753"
    line2 = "2 00005  34.2682 348.7242 1859667 331.7664  19.3264 10.82419157413667"

    return TLE.from_tle(line1, line2, name)


def test_propagation(init_tle_vgd):
    # init TLE
    tle = init_tle_vgd

    # init SGP4
    sgp4 = SGP4Propagator()

    #  Propagate 3 days into future
    time = tle.epoch + 3.0 * u.day
    rv_gcrs = sgp4.propagate(tle, time)

    # print(rv_gcrs)

    # Generate truth values
    # Vallado IAU-76/FK5 - pg.234
    v_j2000_true = CartesianDifferential(
        [-2.233348094, -4.110136162, -3.157394074], unit=u.km / u.s
    )
    r_j2000_true = CartesianRepresentation(
        [-9059.9413786, 4659.6972000, 813.9588875], unit=u.km
    )

    rv_j2000_true = J2000(
        r_j2000_true.with_differentials(v_j2000_true),
        obstime=time,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    rv_gcrs_true = SkyCoord(
        rv_j2000_true.transform_to(GCRS(obstime=time)),
        representation_type="cartesian",
        differential_type="cartesian",
    )

    r_diff, v_diff = rv_diff(rv_gcrs, rv_gcrs_true)

    # print(r_diff.to(u.mm))
    # print(v_diff.to(u.mm / u.s))

    assert r_diff.to(u.mm) < 1700 * u.mm
    assert v_diff.to(u.mm / u.s) < 0.88 * u.mm / u.s


def rv_diff(rv, rv_true):
    """Computes pos and vel diff norm."""
    r_diff = (
        rv.cartesian.without_differentials() - rv_true.cartesian.without_differentials()
    )

    v_diff = rv.velocity - rv_true.velocity

    return r_diff.norm(), v_diff.norm()


def test_decay(init_tle_decay):
    """Tests satellite altitude below threshold - satellite has decayed.
    - should raise `SGP4SatDecayedError`."""
    with pytest.raises(SGP4SatDecayedError):
        # init TLE
        tle = init_tle_decay

        #  Propagate 3 days into future
        time = tle.epoch + 30.0 * u.day
        SGP4Propagator.propagate(tle, time)


def test_crashed(init_tle_decay):
    """Tests satellite altitude below earth radius - satellite has crashed.
    - should raise `SGP4GeneralError`."""
    with pytest.raises(SGP4GeneralError):
        # init TLE
        tle = init_tle_decay

        #  Propagate 3 days into future
        time = tle.epoch + 300.0 * u.day
        SGP4Propagator.propagate(tle, time)


def test_gen_trajectory(init_tle_vgd):
    # init TLE
    tle = init_tle_vgd
    # init SGP4
    sgp4 = SGP4Propagator(stepsize=120 * u.s)

    #  Generate trajectory
    interval = TimeInterval(tle.epoch, 0.223456789 * u.day)
    traj = sgp4.gen_trajectory(tle, interval)

    # Generate true trajectory and compute max pos & vel error
    r_diff_max = 0 * u.mm
    v_diff_max = 0 * u.mm / u.s
    for time in traj.coord_list.obstime:
        rv = traj(time)
        rv_true = SGP4Propagator.propagate(tle, time)
        r_diff, v_diff = rv_diff(rv, rv_true)
        if r_diff > r_diff_max:
            r_diff_max = r_diff
        if v_diff > v_diff_max:
            v_diff_max = v_diff

    # print(r_diff_max.to(u.mm))
    # print(v_diff_max.to(u.mm / u.s))

    assert r_diff_max.to(u.mm) < 5e-6 * u.mm
    assert v_diff_max.to(u.mm / u.s) < 4e-9 * u.mm / u.s



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Coordinate systems and frames defined by `satmad`.

"""
import erfa
import numpy as np
from astropy import units as u
from astropy.coordinates import (
    GCRS,
    ICRS,
    ITRS,
    AffineTransform,
    BaseCoordinateFrame,
    CartesianDifferential,
    DynamicMatrixTransform,
    StaticMatrixTransform,
    TimeAttribute,
    frame_transform_graph,
    get_body_barycentric,
    get_body_barycentric_posvel,
)
from astropy.coordinates import representation as r
from astropy.coordinates.builtin_frames.utils import (
    DEFAULT_OBSTIME,
    get_jd12,
    get_polar_motion,
)

from satmad.core.celestial_bodies import MOON

_w = np.array([0, 0, 7.292115e-5]) / u.s
"""Nominal mean angular velocity of the Earth [rad/s] as per GRS 80.
(see IERS TN 36)"""

_FRAME_BIAS_MATRIX = np.array(
    [
        [0.9999999999999942, 7.078279477857338e-8, -8.056217380986972e-8],
        [-7.078279744199198e-8, 0.9999999999999969, -3.306040883980552e-8],
        [8.056217146976134e-8, 3.3060414542221364e-8, 0.9999999999999962],
    ]
)
"""This is the fixed 3x3 frame bias matrix that does the conversion between the
 J2000 and GCRS frames.
"""


class CelestialBodyCRS(BaseCoordinateFrame):
    """
    A coordinate frame in the generic Celestial Reference System (CRS). This CRS is
    derived from ICRS by simply carrying the origin to the origin of the celestial body.
    A specific example is GCRS, where the origin is at the centre of the Earth.

    """

    obstime = TimeAttribute(default=DEFAULT_OBSTIME)

    default_representation = r.CartesianRepresentation
    default_differential = r.CartesianDifferential

    def __new__(cls, *args, **kwargs):
        frame_transform_graph.transform(AffineTransform, ICRS, cls)(icrs_to_cb_crs)
        frame_transform_graph.transform(AffineTransform, cls, ICRS)(cb_crs_to_icrs)
        return super().__new__(cls)


_NEED_ORIGIN_HINT = (
    "The input {0} coordinates do not have length units. This "
    "probably means you created coordinates with lat/lon but "
    "no distance.  Heliocentric<->ICRS transforms cannot "
    "function in this case because there is an origin shift."
)


def cb_crs_to_icrs(cb_crs_coord, icrs_frame):
    """Conversion from Celestial Reference System of a Central Body to ICRS."""

    if not u.m.is_equivalent(cb_crs_coord.cartesian.x.unit):
        raise u.UnitsError(_NEED_ORIGIN_HINT.format(cb_crs_coord.__class__.__name__))

    if cb_crs_coord.data.differentials:
        # Calculate the barycentric position and velocity (of a solar system body).
        # Uses default ephemeris
        r_icrs, v_icrs = get_body_barycentric_posvel(
            cb_crs_coord.body.name, cb_crs_coord.obstime, ephemeris="builtin"
        )

        v_icrs = CartesianDifferential.from_cartesian(v_icrs)

        # Prepare final coord vector with velocity
        icrs_coord = r_icrs.with_differentials(v_icrs)
    else:
        # Calculate the barycentric position ONLY (of a solar system body).
        # Uses default ephemeris. This is faster than the one above with velocities for
        # JPL ephemerides.
        icrs_coord = get_body_barycentric(
            cb_crs_coord.body.name, cb_crs_coord.obstime, ephemeris="builtin"
        )

    # Return transformation matrix (None) and translation vector (with velocities)
    return None, icrs_coord


def icrs_to_cb_crs(icrs_coord, cb_crs_frame):
    """Conversion from ICRS to Celestial Reference System of a Central Body."""

    if not u.m.is_equivalent(icrs_coord.cartesian.x.unit):
        raise u.UnitsError(_NEED_ORIGIN_HINT.format(icrs_coord.__class__.__name__))

    if icrs_coord.data.differentials:
        # Calculate the barycentric position and velocity (of a solar system body).
        # Uses default ephemeris
        r_icrs, v_icrs = get_body_barycentric_posvel(
            cb_crs_frame.body.name, cb_crs_frame.obstime, ephemeris="builtin"
        )

        v_icrs = CartesianDifferential.from_cartesian(v_icrs)

        # Prepare final coord vector with velocity
        cb_crs_coord = (-r_icrs).with_differentials(-v_icrs)
    else:
        # Calculate the barycentric position ONLY (of a solar system body).
        # Uses default ephemeris. This is faster than the one above with velocities for
        # JPL ephemerides.
        cb_crs_coord = -get_body_barycentric(
            cb_crs_frame.body.name, cb_crs_frame.obstime, ephemeris="builtin"
        )

    # Return transformation matrix (None) and translation vector (with velocities)
    return None, cb_crs_coord


class MoonCRS(CelestialBodyCRS):
    """Moon Celestial Reference System. This is simply the ICRS shifted to the
    centre of the Moon."""

    body = MOON


# ******  Mean Pole and Equinox at J2000.0 Reference System (J2000) ******


class J2000(BaseCoordinateFrame):
    """
    A coordinate or frame in the Mean Pole and Equinox at J2000.0 Reference
    System (J2000).

    This coordinate frame is similar to GCRS but rotated by the frame bias.
    This rotation is applicable only to the
    equinox based approach, and is only an approximation. The difference
    betweenGCRS and J2000 is less than 1m for
    the Low Earth Orbit, therefore these two can be used interchangeably with
    a small error.

    This frame is for legacy applications that store data in J2000. GCRS
    should be used wherever possible.

    References
    ----------
    The definitions and conversions are from IERS Conventions 2010, Chapter 5
    [TCF1]_.

    """

    obstime = TimeAttribute(default=DEFAULT_OBSTIME)

    default_representation = r.CartesianRepresentation
    default_differential = r.CartesianDifferential


# *********** Terrestrial Intermediate Reference System (TIRS) definition
# and conversions. ***********


class TIRS(BaseCoordinateFrame):
    """
    A coordinate or frame in the Terrestrial Intermediate Reference System
    (TIRS).

    References
    ----------
    The definitions and conversions are from IERS Conventions 2010, Chapter 5
    [TCF1]_.

    """

    default_representation = r.CartesianRepresentation
    default_differential = r.CartesianDifferential

    obstime = TimeAttribute(default=DEFAULT_OBSTIME)


def _gmst82_angle(obstime):
    """
    Universal Time to Greenwich mean sidereal time (IAU 1982 model).

    Parameters
    ----------
    obstime : Time
        time at which the polar motion should be calculated.
    Returns
    -------
    float
        Greenwich mean sidereal time (radians)
    """
    # Get GMST82 angle in rad
    # noinspection PyArgumentList
    gmst82 = erfa.gmst82(*get_jd12(obstime, "ut1")) * u.rad

    return gmst82


def _polar_mot_matrix(obstime):
    """
    Form the matrix of polar motion for a given date, IAU 2000.

    The matrix operates in the sense V(TRS) = rpom * V(CIP), meaning that
    it is the final rotation when computing the
    pointing direction to a celestial source.

    Parameters
    ----------
    obstime : Time
        time at which the polar motion should be calculated.
    Returns
    -------
        3x3 rotation matrix due to polar motion
    """
    # compute the polar motion p-matrix
    xp, yp = get_polar_motion(obstime)
    # noinspection PyArgumentList
    sp = erfa.sp00(*get_jd12(obstime, "tt"))
    polar_mot_mat = erfa.pom00(xp, yp, sp)

    return polar_mot_mat


@frame_transform_graph.transform(DynamicMatrixTransform, TIRS, ITRS)
def tirs_to_itrs(tirs_coord, itrs_frame):
    """Dynamic conversion matrix (Polar Motion) from TIRS to ITRS."""
    tirs_to_itrs_mat = _polar_mot_matrix(tirs_coord.obstime)

    return tirs_to_itrs_mat


@frame_transform_graph.transform(DynamicMatrixTransform, ITRS, TIRS)
def itrs_to_tirs(itrs_coord, tirs_frame):
    """Dynamic conversion matrix (Polar Motion) from ITRS to TIRS."""
    itrs_to_tirs_mat = _polar_mot_matrix(itrs_coord.obstime).transpose()

    return itrs_to_tirs_mat


@frame_transform_graph.transform(StaticMatrixTransform, J2000, GCRS)
def j2000_to_gcrs():
    """Constant conversion matrix (Frame Bias) from J2000 to GCRS."""
    return _FRAME_BIAS_MATRIX


@frame_transform_graph.transform(StaticMatrixTransform, GCRS, J2000)
def gcrs_to_j2000():
    """Constant conversion matrix (Frame Bias) from GCRS to J2000."""
    return _FRAME_BIAS_MATRIX.transpose()



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
satmad
=========
Satellite Mission Analysis and Design in Python.
"""
__version__ = "0.1"



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Force models generally for use with the numerical propagators.

"""
import numpy as np


def two_body_accel(r, mu):
    r"""
    Two-body acceleration.

    The input vectors as well as the output acceleration is in in the inertial frame
    of the central body. For the Earth this is GCRS.

    The acceleration value is given by:

    .. math::

        \ddot{\vec{r}} = - \dfrac{\mu}{r^3} \vec{r}

    where :math:`\ddot{\vec{r}}` is the inertial acceleration,
    :math:`\vec{r}` is the position vector (with :math:`r` its norm).
    :math:`\mu` is equal to the constant :math:`GM` (Gravitational Constant times
    the Mass of the main attracting body.)

    .. note::
        This method is used in the Scipy ODE solver, therefore it does not accept
        `Quantity` inputs. Care must be taken with the units.


    Parameters
    ----------
    r : ndarray
        inertial position vector with 3 elements (km)
    mu : float
        GM value (:math:`km^3 / s^2`)

    Returns
    -------
    float
        inertial acceleration (:math:`km / s^2`)

    """
    r_norm = np.linalg.norm(r)
    a = -mu / (r_norm ** 3) * r
    return a


def two_body_energy(r, v, mu):
    r"""Two-body specific energy.

    The input vectors are in in the inertial frame of the central body.
    For the Earth this is GCRS.

    The specific energy value is given by:

    .. math::

        \varepsilon = \dfrac{v^2}{2} - \dfrac{\mu}{r}

    with :math:`v` is the norm of the velocity vector, :math:`r` is the norm of the
    position vector and :math:`\mu` is equal to the constant :math:`GM`
    (Gravitational Constant times the Mass of the main attracting body).

    This is the formula for the specific energy, or the sum of
    potential and kinetic energy *per unit mass*.

    Parameters
    ----------
    r : ndarray or Quantity
        inertial position vector with 3 elements (km)
    v : ndarray or Quantity
        inertial position vector with 3 elements (km/s)
    mu : float or Quantity
        GM value (:math:`km^3 / s^2`)

    Returns
    -------
    sp_energy : float or Quantity
        specific energy (:math:`km^2 / s^2`)
    """
    r_norm = np.linalg.norm(r)
    v_norm = np.linalg.norm(v)
    sp_energy = 0.5 * v_norm ** 2 - mu / r_norm

    return sp_energy



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Definition of a Central Body, which is a celestial body around which
an object can rotate.

"""
from astropy import units as u
from astropy.coordinates import frame_transform_graph


class CelestialBodyEllipsoid:
    """Defines an ellipsoid, the shape of most celestial bodies.

    Parameters
    ----------
    name : str
        Name of the Ellipsoid
    re : Quantity
        Equatorial Radius or Semimajor Axis of the Ellipsoid. (km)
    inv_f : Quantity
        Inverse flattening (:math:`1/f`). (dimensionless)
    j2 : Quantity
        Dynamical form factor
    om : Quantity
        Nominal mean angular velocity of the celestial body
    mu : Quantity
        `GM` value (Gravitational constant) of the Celestial body
    """

    j2 = None
    mu = None

    om = None

    @u.quantity_input(
        re=u.km,
        inv_f=u.dimensionless_unscaled,
        j2=u.dimensionless_unscaled,
        mu=u.m ** 3 / u.s ** 2,
        om=u.rad / u.s,
    )
    def __init__(self, name, re, inv_f, **kwargs):
        self.name = name

        self.re = re
        self.inv_f = inv_f

        self.j2 = kwargs.get("j2")
        self.mu = kwargs.get("mu")

        self.om = kwargs.get("om")


class CelestialBody:
    """
    Container class for Celestial Bodies and related information.

    Parameters
    ----------
    name : str
        Name of the body
    info : str
        Info on the constants etc.
    mu : Quantity
        `GM` value (Gravitational constant) of the Celestial body
    inert_coord : str
        Default Inertial Coordinate to run the propagations (e.g. "gcrs" for Earth)
    body_fixed_coord : str
        Default Central Body Fixed Coordinate to run the propagations (e.g. "itrs" for
        Earth)
    """

    @u.quantity_input(mu=u.km ** 3 / u.s ** 2)
    def __init__(self, name, info, mu, **kwargs):
        self.name = name
        self.info = info

        self.mu = mu

        self.ellipsoid = kwargs.get("ellipsoid")

        self.inert_coord = kwargs.get("inert_coord")
        self.body_fixed_coord = kwargs.get("body_fixed_coord")

    def __str__(self) -> str:
        out = f"Name: {self.name}\n"

        out += f"{self.info}\n"

        out += f"mu: {self.mu}\n"

        out += f"Default inertial coord: {self.inert_coord}\n"
        out += f"Default body fixed coord: {self.body_fixed_coord}\n"

        return out

    @property
    def inert_coord_frame(self):
        """Gets the underlying inertial coordinate frame class (e.g. GCRS for Earth)
        or None if no such class exists."""
        return frame_transform_graph.lookup_name(self.inert_coord)

    @property
    def body_fixed_coord_frame(self):
        """Gets the underlying body fixed coordinate frame class (e.g. ITRS for Earth)
        or None if no such class exists."""
        return frame_transform_graph.lookup_name(self.body_fixed_coord)



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Two-Line-Elements to represent satellites in Earth orbit.

"""
import numpy as np
from astropy import units as u
from astropy.time import Time
from astropy.units import Quantity
from sgp4.earth_gravity import wgs72
from sgp4.exporter import export_tle
from sgp4.model import WGS72, Satrec

# precompile the unit as constant
from satmad.core.celestial_bodies import EARTH_ELLIPSOID_WGS84

_DAY_TO_SEC = (1.0 * u.day).to(u.s)


class TLE:
    """
    A two-line element set (TLE) is a data format encoding a list of TEME
    (True Equator, Mean Equinox) mean orbital elements
    of an Earth-orbiting object for a given point in time, called the Epoch Time.

    These orbital elements are solely for use with the SGP4 propagator due to the
    analytical orbit theory used in its derivation.

    See the `TLE page in Wikipedia <https://en.wikipedia.org/wiki/Two-line_element_set>`_
    or `NASA definition <https://spaceflight.nasa.gov/realdata/sightings/SSapplications/Post/JavaSSOP/SSOP_Help/tle_def.html>`_
    for more information.

    The TLEs are usually generated by external sources and are used to propagate
    the orbits with the initial condition encapsulated by the TLE.

    Parameters
    ----------
    epoch : Time
        Epoch Time corresponding to the orbital elements (nominally very near
        the time of true ascending node passage)
    inclination : float or Quantity
        TEME mean inclination of the orbit [rad]
    raan : float or Quantity
        TEME mean right ascension of ascending node (RAAN) of the orbit [rad]
    eccentricity : float or Quantity
        mean eccentricity of the orbit [dimensionless]
    arg_perigee : float or Quantity
        TEME mean argument of perigee [rad]
    mean_anomaly : float or Quantity
        mean anomaly of the orbit [rad]
    mean_motion : float or Quantity
        mean motion of the orbit [rad/sec]
    bstar : float or Quantity
        sgp4 type drag coefficient [1 / earth radius] (see TLE class documentation)
    n_dot : float or Quantity
        First time derivative of the mean motion or
        Ballistic Coefficient [revs/day] (see TLE class documentation)
    n_dotdot : float or Quantity
        Second time derivative of the mean motion (see TLE class documentation)
    name : str
        Common name of the satellite
    intl_designator : str
        international designator on card 1 (up to 8 characters) (see class definition)
    sat_num : int
        satellite catalog number (see TLE class documentation)
    classification : str
        Classification (`U` for Unclassified, `C` for Classified, `S` for Secret)
    rev_nr : int
        Revolution number of the object at Epoch Time [revolutions]
    el_nr : int
        Element set number. Incremented when a new TLE is generated for this object.
    """

    # Gravitational constants(defaults to WGS72)
    _grav_model = WGS72
    _mu = wgs72.mu * (u.km ** 3 / u.s ** 2)  # in km3 / s2
    _earth_radius = wgs72.radiusearthkm * u.km  # km
    _j2 = wgs72.j2

    # Hardcoded defaults
    _ops_mode = "i"

    def __init__(
        self,
        epoch,
        inclination,
        raan,
        eccentricity,
        arg_perigee,
        mean_anomaly,
        mean_motion,
        bstar,
        n_dot,
        n_dotdot=0.0,
        name="NONAME",
        intl_designator="12345A",
        sat_num=99999,
        classification="U",
        rev_nr=0,
        el_nr=1,
    ):
        # init internal satrec object
        self._satrec: Satrec = Satrec()

        # recreate the epoch for use with Satrec
        self.__create_epoch_composite(epoch)
        # epoch_yydd = self._satrec.epochyr * 1000 + self._satrec.epochdays
        epoch_satrec = epoch.jd - 2433281.5

        # check inclination range
        inclination_val = _force_angles_to_rad(inclination)
        if not 0 <= inclination_val < np.pi:
            raise ValueError(
                f"Given argument ({inclination}) is an invalid "
                f"Inclination value, "
                f"only values in range 0 <= i < PI are allowed."
            )

        # check mean motion
        if isinstance(mean_motion, Quantity):
            mean_motion = mean_motion.to(
                u.rad / u.s, equivalencies=u.dimensionless_angles()
            ).value

        # check eccentricity as Quantity and reduce to value
        if isinstance(eccentricity, Quantity):
            eccentricity = eccentricity.value

        # fill the Satrec object
        self._satrec.sgp4init(
            TLE._grav_model,
            TLE._ops_mode,
            sat_num,
            epoch_satrec,
            bstar,
            n_dot,
            n_dotdot,
            eccentricity,
            _force_angle_range(arg_perigee),
            inclination_val,
            _force_angle_range(mean_anomaly),
            mean_motion * 60,  # convert to seconds
            _force_angle_range(raan),
        )

        # fill time with precise Time information
        self._epoch = epoch.utc.replicate(format="isot")
        self._satrec.jdsatepoch = self._epoch.jd1
        self._satrec.jdsatepochF = self._epoch.jd2

        # load other TLE info
        self._satrec.classification = classification
        self._satrec.intldesg = intl_designator
        self._satrec.revnum = rev_nr
        self._satrec.ephtype = "0"  # always 0 in distributed ephemeris
        self._satrec.elnum = el_nr

        # load TLE name
        self._name = name

    @classmethod
    def from_tle(cls, line1, line2, name="No Name"):
        """
        Initialises the TLE from two strings.

        Parameters
        ----------
        line1 : str
            First line of the TLE
        line2 : str
            Second line of the TLE
        name : str
            Name of the object / satellite

        Returns
        -------
        TLE
            `TLE` object initialised with the two line input.
        """
        # create object without calling `__init__`
        tle = cls.__new__(cls)

        # init Satrec object with TLE strings
        tle._satrec = Satrec.twoline2rv(line1, line2, whichconst=TLE._grav_model)

        # Fill epoch in Time
        tle._epoch = Time(
            tle._satrec.jdsatepoch, tle._satrec.jdsatepochF, scale="utc", format="jd"
        )

        # load TLE name
        tle._name = name

        return tle

    @classmethod
    def init_geo(
        cls,
        epoch,
        longitude,
        name="No Name",
        intl_designator="12345A",
        sat_num=99999,
        classification="U",
        rev_nr=0,
        el_nr=1,
    ):
        """
        Initialises a geostationary satellite TLE.

        Due to the nature of the Earth's geopotential, the orbit may drift off
        by some degrees within weeks.

        Parameters
        ----------
        epoch : Time
            Epoch Time corresponding to the orbital elements (nominally very near
            the time of true ascending node passage)
        longitude : float or Quantity
            initial longitude of the satellite
        name : str
            Common name of the satellite
        intl_designator : str
            international designator on card 1 (up to 8 characters) (see class definition)
        sat_num : int
            satellite catalog number (see TLE class documentation)
        classification : str
            Classification (`U` for Unclassified, `C` for Classified, `S` for Secret)
        rev_nr : int
            Revolution number of the object at Epoch Time [revolutions]
        el_nr : int
            Element set number. Incremented when a new TLE is generated for this object.

        Returns
        -------
        TLE
            `TLE` object initialised with the required GEO parameters.
        """
        # init GEO specific values - period is one sidereal day
        mean_motion = 2 * np.pi / (1.0 * u.sday).to_value(u.s)
        raan = epoch.sidereal_time("mean", longitude)

        bstar = 0  # no drag
        n_dot = 0  # mean motion assumed constant

        # init standard (near) circular orbit
        inclination = 0.0 * u.deg
        eccentricity = 1e-9  # cannot be zero
        arg_perigee = 0.0 * u.deg
        mean_anomaly = 0.0 * u.deg

        tle = TLE(
            epoch,
            inclination,
            raan,
            eccentricity,
            arg_perigee,
            mean_anomaly,
            mean_motion,
            bstar,
            n_dot,
            n_dotdot=0.0,
            name=name,
            intl_designator=intl_designator,
            sat_num=sat_num,
            classification=classification,
            rev_nr=rev_nr,
            el_nr=el_nr,
        )

        return tle

    @classmethod
    def init_sso(
        cls,
        epoch,
        altitude,
        ltan,
        eccentricity=1e-9,
        arg_perigee=0 * u.deg,
        mean_anomaly=0.0 * u.deg,
        earth_radius=EARTH_ELLIPSOID_WGS84.re,
        bstar=0,
        name="No Name",
        intl_designator="12345A",
        sat_num=99999,
        classification="U",
        rev_nr=0,
        el_nr=1,
    ):
        """
        Initialises a sun-synchronous Earth orbit TLE.

        Sets the inclination in accordance with the target node rotation rate that
        satisfies sun-synchronous conditions up to J2. Sets the RAAN in accordance
        with the requested Local Time of the Ascending Node (LTAN).

        Target node rotation rate is 360 deg / 365.2421897 days (See
        Fundamentals of Astrodynamics Vallado 4th ed pg. 863)

        Parameters
        ----------
        epoch : Time
            Epoch Time corresponding to the orbital elements (nominally very near
            the time of true ascending node passage)
        altitude : float or Quantity
            altitude above `earth_radius` around Equator [km]
        ltan : float or Quantity
            Local Time of the Ascending Node defined as [0, 24) or a hour angle
        earth_radius :  float or Quantity
            Earth radius at equator - defines the semimajor axis together with the
            `altitude` parameter [km]
        eccentricity : float
            mean eccentricity of the orbit [dimensionless]
        arg_perigee : float or Quantity
            TEME mean argument of perigee [rad]
        mean_anomaly : float or Quantity
            mean anomaly of the orbit [rad]
        bstar : float or Quantity
            sgp4 type drag coefficient [1 / earth radius] (see TLE class documentation)
        name : str
            Common name of the satellite
        intl_designator : str
            international designator on card 1 (up to 8 characters) (see class definition)
        sat_num : int
            satellite catalog number (see TLE class documentation)
        classification : str
            Classification (`U` for Unclassified, `C` for Classified, `S` for Secret)
        rev_nr : int
            Revolution number of the object at Epoch Time [revolutions]
        el_nr : int
            Element set number. Incremented when a new TLE is generated for this object.

        Returns
        -------
        TLE
            `TLE` object initialised with the required SSO parameters.
        """
        sma = altitude + earth_radius
        # mean motion - convert to float
        mean_motion = np.sqrt(TLE._mu / sma ** 3).value

        # mean motion assumed constant
        n_dot = 0

        if eccentricity < 1e-9:
            eccentricity = 1e-9  # cannot be zero or negative

        # target omega_dot value for sun sync
        om_dot_sun_sync = 360.0 / 365.2421897 * u.deg / u.day

        # use consistent set of units (defaults from TLE object)
        inclination = (
            np.arccos(
                (
                    (-2 * sma ** 3.5 * om_dot_sun_sync * (1 - eccentricity ** 2) ** 2)
                    / (3 * TLE._earth_radius ** 2 * TLE._j2 * np.sqrt(TLE._mu))
                )
                .decompose()
                .to_value(u.rad)
            )
            * u.rad
        )

        # compute RAAN
        raan = (
            epoch.sidereal_time("mean", 0)
            + ((0.5 - epoch.jd2) + ltan / 24) * 360 * u.deg
        )

        raan = _force_angle_range(raan)

        tle = TLE(
            epoch,
            inclination,
            raan,
            eccentricity,
            arg_perigee,
            mean_anomaly,
            mean_motion,
            bstar,
            n_dot,
            n_dotdot=0.0,
            name=name,
            intl_designator=intl_designator,
            sat_num=sat_num,
            classification=classification,
            rev_nr=rev_nr,
            el_nr=el_nr,
        )

        return tle

    def __str__(self):
        """
        Exports the TLE as a Three-Line string, with name, line1 and line2.

        A sample output looks like this::

            ISS (ZARYA)
            1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927
            2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537

        Returns
        -------
        str
            string containing line 1, line 2 and name of the satellite
        """
        line1, line2 = export_tle(self._satrec)

        txt = ""
        if self._name is not None and self._name.strip():
            txt += self._name + "\n"

        txt += line1 + "\n" + line2 + "\n"

        return txt

    def export_tle(self):
        """
        Exports the TLE to line1 and line2.

        A sample output looks like this::

            line1 = "1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927"
            line2 = "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537"

        Returns
        -------
        line1, line2 : (str, str)
            strings containing line 1, line 2
        """
        line1, line2 = export_tle(self._satrec)

        return line1, line2

    @property
    def period(self) -> Quantity:
        """Computes the satellite period [sec]."""
        return (2 * np.pi / self.mean_motion).to(
            u.s, equivalencies=u.dimensionless_angles()
        )

    @property
    def sm_axis(self) -> Quantity:
        """Computes the semimajor axis [km].

        Note that the :math:`mu` value used in the computation of the semimajor axis
        is that of WGS72 for consistency with the orbital elements definition.
        """
        n = self.mean_motion.to_value(u.rad / u.s)
        return np.power(TLE._mu.value / n / n, 1 / 3) * u.km

    @property
    def node_rotation_rate(self):
        """
        Node (or orbital plane) rotation rate due to J2.

        Note that, WGS72 constants for J2 and Earth Radius (R_E) are used in
        the computation for consistency with the orbital elements definition.

        Returns
        -------
        Quantity
            Node rotation rate (deg/day)
        """
        n = self.mean_motion
        r_e = TLE._earth_radius
        e = self.eccentricity
        p = self.sm_axis * (1 - e * e)
        j2 = TLE._j2
        i = self.inclination

        node_rot = -3 * n * r_e * r_e * j2 / (2 * p * p) * np.cos(i)

        return node_rot.to(u.deg / u.day)

    @property
    def argp_rotation_rate(self):
        """
        Argument of Perigee (or eccentricity vector) rotation rate due to J2.

        Note that, WGS72 constants for J2 and Earth Radius (R_E) are used in
        the computation for consistency with the orbital elements definition.

        Returns
        -------
        Quantity
            Arg of Perigee rotation rate (deg/day)
        """
        n = self.mean_motion
        r_e = TLE._earth_radius
        e = self.eccentricity
        p = self.sm_axis * (1 - e * e)
        j2 = TLE._j2
        sin_i = np.sin(self.inclination)

        argp_rot = (
            3.0 * n * r_e * r_e * j2 / (4.0 * p * p) * (4.0 - 5.0 * sin_i * sin_i)
        )

        return argp_rot.to(u.deg / u.day)

    @property
    def satrec(self) -> Satrec:
        """Returns the underlying `Satrec` object. This is for internal use only,
        for use with the SGP4 propagator."""
        return self._satrec

    @property
    def epoch(self) -> Time:
        """Returns the epoch time associated with the orbital parameters."""
        return self._epoch

    @epoch.setter
    def epoch(self, epoch: Time):
        self._epoch = epoch.replicate()

        self._satrec.jdsatepoch = self._epoch.jd1
        self._satrec.jdsatepochF = self._epoch.jd2

        # recreate the epoch composite
        self.__create_epoch_composite(epoch)

    @property
    def bstar(self) -> float:
        """Returns the sgp4 type drag coefficient [1 / earth radius]."""
        return self._satrec.bstar

    @property
    def n_dot(self) -> float:
        """
        Gets and sets the first time derivative of the mean motion or
        Ballistic Coefficient [revs/day].

        Absolute value of the first time derivative of the mean motion
        can only be less than 1.0 (exclusive). Raises a `ValueError` otherwise.
        """
        return self._satrec.ndot

    @n_dot.setter
    def n_dot(self, n_dot):
        if abs(n_dot) < 1.0:
            self._satrec.ndot = n_dot
        else:
            raise ValueError(
                f"Given argument ({n_dot}) is an invalid derivative of mean motion, "
                f"only absolute values less than 1.0 (exclusive) are allowed."
            )

    @property
    def n_dotdot(self) -> float:
        """Returns the Second time derivative of the mean motion."""
        return self._satrec.nddot

    @property
    def inclination(self) -> Quantity:
        """Gets and sets the TEME mean inclination of the orbit [rad].

        Inclination should be in range 0 <= om < PI.
        Raises a `ValueError` otherwise.
        """
        return self._satrec.inclo * u.rad

    @inclination.setter
    def inclination(self, i):
        """
        Parameters
        ----------
        i : float or Quantity
        """
        i_val = _force_angles_to_rad(i)

        if 0 <= i_val < np.pi:
            self._satrec.inclo = i_val
        else:
            raise ValueError(
                f"Given argument ({i}) is an invalid "
                f"Inclination value, "
                f"only values in range 0 <= i < PI are allowed."
            )

    @property
    def raan(self) -> Quantity:
        """Gets and sets the TEME mean right ascension of ascending node (RAAN)
        of the orbit [rad].

        RAAN should be in range 0 <= om < 2*PI.
        Raises a `ValueError` otherwise."""
        return self._satrec.nodeo * u.rad

    @raan.setter
    def raan(self, om):
        """
        Parameters
        ----------
        om : float or Quantity
        """
        om_val = _force_angle_range(om)

        if 0 <= om_val < 2 * np.pi:
            self._satrec.nodeo = om_val
        else:
            raise ValueError(
                f"Given argument ({om}) is an invalid "
                f"RAAN value, "
                f"only values in range 0 <= om < 2*PI are allowed."
            )

    @property
    def eccentricity(self) -> float:
        """Gets and sets the mean eccentricity of the orbit.

        Input can be a Quantity (dimensionless) or float. Getter value is a float.

        Eccentricity should be in range 0 <= e < 1.0.
        Raises a `ValueError` otherwise."""
        return self._satrec.ecco

    @eccentricity.setter
    def eccentricity(self, e):

        # check eccentricity as Quantity and reduce to value
        if isinstance(e, Quantity):
            e = e.value

        if 0 <= e < 1.0:
            self._satrec.ecco = e
        else:
            raise ValueError(
                f"Given argument ({e}) is an invalid "
                f"Eccentricity value, "
                f"only values in range 0 <= e < 1.0 are allowed."
            )

    @property
    def arg_perigee(self) -> Quantity:
        """Gets and sets the TEME mean argument of perigee [rad].

        Argument of perigee should be in range 0 <= argp < 2*PI.
        Raises a `ValueError` otherwise.
        """
        return self._satrec.argpo * u.rad

    @arg_perigee.setter
    def arg_perigee(self, argp):
        """
        Parameters
        ----------
        argp : float or Quantity
        """
        argp_val = _force_angle_range(argp)

        if 0 <= argp_val < 2 * np.pi:
            self._satrec.argpo = argp_val
        else:
            raise ValueError(
                f"Given argument ({argp}) is an invalid "
                f"Argument of Perigee value, "
                f"only values in range 0 <= om < 2*PI are allowed."
            )

    @property
    def mean_anomaly(self) -> Quantity:
        """Gets and sets the mean anomaly of the orbit [rad].

        Mean Anomaly should be in range 0 <= m_anom < 2*PI.
        Raises a `ValueError` otherwise.
        """
        return self._satrec.mo * u.rad

    @mean_anomaly.setter
    def mean_anomaly(self, m_anom):
        """
        Parameters
        ----------
        m_anom : float or Quantity
        """
        m_anom_val = _force_angles_to_rad(m_anom)
        if 0 <= m_anom_val < 2 * np.pi:
            self._satrec.mo = m_anom_val
        else:
            raise ValueError(
                f"Given argument ({m_anom}) is an invalid "
                f"Mean Anomaly value, "
                f"only values in range 0 <= m_anom < 2*PI are allowed."
            )

    @property
    def mean_motion(self) -> Quantity:
        """Gets and sets the mean motion of the orbit [rad/sec].

        Setter input can be in float [rad/sec] or Quantity. Getter output is in
        Quantity.

        Mean Motion in revs/day should be in range 0 < n <= 17.0.
        Raises a `ValueError` otherwise.
        """
        return self._satrec.no_kozai / 60.0 * u.rad / u.s

    @mean_motion.setter
    def mean_motion(self, n):

        # check mean motion as Quantity
        if isinstance(n, Quantity):
            n = n.to(u.rad / u.s, equivalencies=u.dimensionless_angles()).value

        no = _DAY_TO_SEC.to_value() * n / (2 * np.pi)
        if 0 < no <= 17.0:
            self._satrec.no_kozai = n * 60
        else:
            raise ValueError(
                f"Given argument ({n}, converted to {no}) is an invalid "
                f"Mean Motion value, "
                f"only values in range 0 < n <= 17.0 are allowed."
            )

    @property
    def name(self) -> str:
        """Gets and sets the common name of the satellite.

        Satellite name cannot be `None` or empty, "NO NAME" is then assigned
        automatically.
        """
        return self._name

    @name.setter
    def name(self, name):
        if name is None or not name:
            self._name = "NO NAME"
        else:
            self._name = name

    @property
    def sat_number(self) -> int:
        """Returns the satellite catalog number."""
        return self._satrec.satnum

    @property
    def intl_designator(self) -> str:
        """
        Gets or sets the international designator of the satellite.

        International Designator can only be between 0 and 100000
        (exclusive at both ends). Raises a `ValueError` otherwise.
        """
        return self._satrec.intldesg

    @intl_designator.setter
    def intl_designator(self, intl_designator):
        self._satrec.intldesg = intl_designator

    @property
    def classification(self) -> str:
        """
        Gets or sets the classification of the satellite.

        Classification type can only be `U`, `C` or `S`. Raises a `ValueError`
        otherwise.
        """
        return self._satrec.classification

    @classification.setter
    def classification(self, sec_class):
        if sec_class == "U" or sec_class == "C" or sec_class == "S":
            self._satrec.classification = sec_class
        else:
            raise ValueError(
                f"Given argument ({sec_class}) is an invalid classification type, "
                f"only U, C, S are allowed."
            )

    @property
    def rev_nr(self) -> int:
        """
        Gets or sets the Revolution Number of the object at Epoch Time [revolutions].

        Revolution Number can only be between 0 and 100000
        (exclusive at both ends). Raises a `ValueError` otherwise.
        """
        return self._satrec.revnum

    @rev_nr.setter
    def rev_nr(self, rev_nr):
        if 0 < rev_nr < 100000:
            self._satrec.revnum = rev_nr
        else:
            raise ValueError(
                f"Given argument ({rev_nr}) is an invalid Revolution Number, "
                f"only values between 0 and 100000 (exclusive) are allowed."
            )

    @property
    def el_nr(self) -> int:
        """Gets and sets the Element set number.

        Incremented when a new TLE is generated for this object.

        Element set number should be in range 0 <= el_nr < 10000.
        Raises a `ValueError` otherwise.
        """
        return self._satrec.elnum

    @el_nr.setter
    def el_nr(self, el_nr):
        if 0 <= el_nr < 10000:
            self._satrec.elnum = el_nr
        else:
            raise ValueError(
                f"Given argument ({el_nr}) is an invalid Element Set Number, "
                f"only values between 0 and 10000 (exclusive) are allowed."
            )

    def __create_epoch_composite(self, epoch):
        """Fills the `self._satrec.epochyr` and `self._satrec.epochyr` fields
        with the `epoch` input."""
        yydd_str = epoch.utc.to_value("yday", subfmt="date").split(":")
        self._satrec.epochyr = int(yydd_str[0][-2:])
        self._satrec.epochdays = int(yydd_str[1]) + epoch.mjd % 1


def _force_angles_to_rad(angle):
    """
    Forces a `Quantity` or a float to output float in radians.

    Parameters
    ----------
    angle : float or Quantity

    Returns
    -------
    float
        The angle in radians
    """
    if isinstance(angle, Quantity):
        angle_rad = angle.to_value(u.rad)
    else:
        angle_rad = angle

    return angle_rad


def _force_angle_range(angle_rad, min_range=0, max_range=2 * np.pi):
    """
    Forces the angles into a prescribed range.

    Parameters
    ----------
    angle_rad : float or Quantity
        Angle in radians
    min_range : float
        Minimum value allowable
    max_range : float
        Max value allowable

    Returns
    -------
    float
        Angle value forced into the required range
    """

    angle = _force_angles_to_rad(angle_rad)

    while not (min_range <= angle < max_range):
        if max_range < angle:
            angle = angle - 2 * np.pi

        if angle < min_range:
            angle = angle + 2 * np.pi

    return angle



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2021 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Ground Location class tests.

"""
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import (
    ITRS,
    CartesianRepresentation,
    EarthLocation,
    Latitude,
    Longitude,
)
from astropy.time import Time, TimeDelta
from pytest import approx

from satmad.core.celestial_bodies import EARTH, EARTH_ELLIPSOID_WGS84, MOON
from satmad.core.ground_location import GeodeticLocation, GroundLocation


def test_get_body_fixed():
    """Tests the body fixed coords output for multiple times."""
    time = Time("2020-04-10T00:00:00", scale="utc") + np.arange(1, 4) * TimeDelta(
        1, format="jd"
    )

    gnd_loc = GroundLocation(
        10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=EARTH.ellipsoid
    )

    gnd_loc_body_fixed = gnd_loc.to_body_fixed_coords(EARTH, obstime=time)

    r_itrs_true = ITRS(
        CartesianRepresentation(
            np.asarray(
                [
                    [6068714.27712043, 1070078.065267, 1640138.96300037],
                    [6068714.27712043, 1070078.065267, 1640138.96300037],
                    [6068714.27712043, 1070078.065267, 1640138.96300037],
                ]
            ).transpose(),
            unit=u.m,
        ),
        obstime=time,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    # print(gnd_loc_body_fixed)

    assert str(gnd_loc_body_fixed) == str(r_itrs_true)


def test_get_body_fixed_no_coords():
    """Tests the body fixed coords output but the celestial body has no
    body fixed coordinates defined."""
    with pytest.raises(TypeError):

        gnd_loc = GroundLocation(
            10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=MOON.ellipsoid
        )

        gnd_loc.to_body_fixed_coords(MOON)


def test_copy():
    """Tests the initialisation with copy."""
    gnd_loc = GroundLocation(
        10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=EARTH_ELLIPSOID_WGS84
    )

    gnd_loc_copy = GroundLocation(gnd_loc)

    assert gnd_loc == gnd_loc_copy


def test_init_fail_1():
    """Tests the initialisation failure case with cartesian input error."""
    with pytest.raises(TypeError):
        gnd_loc = GroundLocation(
            10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=EARTH_ELLIPSOID_WGS84
        ).to_geocentric

        GroundLocation(gnd_loc)


def test_init_fail_2():
    """Tests the initialisation failure case with cartesian input with no units error."""
    with pytest.raises(TypeError):
        gnd_loc = GroundLocation(
            10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=EARTH_ELLIPSOID_WGS84
        ).to_geocentric()

        # strip unit information
        GroundLocation.from_geocentric(
            gnd_loc[0].to_value(),
            gnd_loc[1].to_value(),
            gnd_loc[2].to_value(),
        )


def test_round_trip():
    """Round trip testing from geodetic to geocentric to geodetic."""
    gd_loc_origin = GeodeticLocation(
        Longitude(10 * u.deg), Latitude(15 * u.deg), 150 * u.m
    )

    # the values are in cartesian coordinates
    gd_loc_copy = GroundLocation(gd_loc_origin).geodetic

    assert gd_loc_origin.lat.to_value(u.deg) == approx(
        gd_loc_copy.lat.to_value(u.deg), abs=1e-14
    )

    assert gd_loc_origin.lon.to_value(u.deg) == approx(
        gd_loc_copy.lon.to_value(u.deg), abs=1e-14
    )

    assert gd_loc_origin.height.to_value(u.m) == approx(
        gd_loc_copy.height.to_value(u.m), abs=1e-8
    )


def test_lat_lon_alt_init():
    """Tests the GroundLocation geodetic init against astropy EarthLocation."""
    gnd_loc_astropy = EarthLocation.from_geodetic(
        10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid="WGS84"
    )

    gnd_loc_gd = GroundLocation.from_geodetic(
        10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=EARTH_ELLIPSOID_WGS84
    )

    gnd_loc = GroundLocation(
        10 * u.deg, 15 * u.deg, 150 * u.m, ellipsoid=EARTH_ELLIPSOID_WGS84
    )

    print(gnd_loc_astropy.geocentric)
    # print(gnd_loc_gd)

    assert gnd_loc_gd == gnd_loc == gnd_loc_astropy


def test_cartesian_init():
    """Tests the GroundLocation cartesian init against astropy EarthLocation."""
    gnd_loc_astropy = EarthLocation.from_geodetic(10 * u.deg, 15 * u.deg, 1500 * u.m)

    gnd_loc_astropy_xyz = gnd_loc_astropy.to_geocentric()

    gnd_loc_gd = GroundLocation.from_geocentric(
        gnd_loc_astropy_xyz[0],
        gnd_loc_astropy_xyz[1],
        gnd_loc_astropy_xyz[2],
        ellipsoid=EARTH_ELLIPSOID_WGS84,
    )

    gnd_loc = GroundLocation(
        gnd_loc_astropy_xyz[0],
        gnd_loc_astropy_xyz[1],
        gnd_loc_astropy_xyz[2],
        ellipsoid=EARTH_ELLIPSOID_WGS84,
    )

    # print(gnd_loc_astropy)
    # print(gnd_loc_gd)

    assert gnd_loc_gd == gnd_loc == gnd_loc_astropy



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Time interval module.

`TimeInterval` class stores time intervals and `TimeIntervalList` class stores
lists of `TimeInterval` objects.
"""

import portion as p
from astropy import units as u
from astropy.time import Time, TimeDelta
from astropy.units import Quantity

_EPS_TIME = 10 * u.ns
"""Allowable time threshold, this much 'out of bounds' is allowed when assuming two
instances of time are *practically* equal. This helps with floating point artifacts
such as round-off errors."""


class TimeInterval:
    """
    Represent and manipulate a single time interval.

    This is a thin wrapper around the  :class:`portion.interval.Interval` class
    from `portion` package (for the Atomic intervals), using Astropy `Time` classes
    under the hood.


    Note that:

    * only the first element of `start_time` and `end_time` objects are used,
      even if they are defined as arrays.

    * `end_time` should be greater than (or later than) `start_time`.
      Zero duration time intervals are not allowed and will raise `ValueError`.

    Parameters
    ----------
    start_time : Time or TimeInterval
        Initial time to mark the start of the interval (only the first
        instance contained in the `Time` object is used). If specified as TimeInterval,
        it is copied into this and `end_time` is ignored.
    end_time : Time, TimeDelta, Quantity or None
        End time or duration to mark the end of the interval
        (only the first instance contained in the `Time` or `TimeDelta` object is used)
        None is acceptable only when `start_time` is defined as a `TimeInterval`
    replicate : bool
        `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False` to
        use a shallow copy to save memory
    start_inclusive : bool
        True if the start of the interval is inclusive (closed), False if exclusive
        (open)
    end_inclusive : bool
        True if the start of the interval is inclusive (closed), False if exclusive
        (open)

    Raises
    ------
    ValueError
        Raised if `end_time` is not an instance of `Time` or `TimeDelta`. Also raised
        if `end_time` <= `start_time`.
    """

    def __init__(
        self,
        start_time,
        end_time,
        replicate=False,
        start_inclusive=True,
        end_inclusive=True,
    ):

        start_interval = None
        end_interval = None
        if isinstance(start_time, TimeInterval):
            # start_time is a valid TimeInterval, just copy its elements
            start_inclusive = start_time.p_interval.left
            end_inclusive = start_time.p_interval.right

            if replicate:
                start_interval = start_time.start.replicate()
                end_interval = start_time.end.replicate()
            else:
                start_interval = start_time.start
                end_interval = start_time.end
        else:
            # Check whether init and end times are scalar or array
            if start_time.isscalar:
                start_tmp = start_time
            else:
                start_tmp = start_time[0]

            if end_time.isscalar:
                end_tmp = end_time
            else:
                end_tmp = end_time[0]

            # Check the type of end_time and fill values accordingly
            if isinstance(end_tmp, TimeDelta):
                # end time is a TimeDelta (duration)

                if replicate:
                    start_interval = start_tmp.replicate()
                    end_interval = start_tmp + end_tmp
                else:
                    start_interval = start_tmp
                    end_interval = start_tmp + end_tmp

            elif isinstance(end_tmp, Quantity):
                # end time is a Quantity, but check whether it is a valid Time Quantity
                if end_tmp.unit.decompose().physical_type == "time":
                    start_interval = start_tmp
                    end_interval = start_tmp + end_tmp

            elif isinstance(end_tmp, Time) and not isinstance(end_tmp, TimeDelta):
                # end time is a Time

                # replicate or shallow copy start and end values - use the first
                # time instances only
                if start_tmp < end_tmp:
                    # end time is later than start time - no probs
                    if replicate:
                        start_interval = start_tmp.replicate()
                        end_interval = end_tmp.replicate()
                    else:
                        start_interval = start_tmp
                        end_interval = end_tmp
                else:
                    # end time is earlier than start time - raise error
                    raise ValueError(
                        f"End time ({end_tmp.isot}) is earlier than start time"
                        f"({start_tmp.isot})"
                    )
            else:
                raise ValueError(
                    f"End time is an instance of {end_tmp.__class__()}, "
                    f"only Time, Quantity (Temporal) or TimeDelta classes are allowed."
                )

        # Initialise the interval
        self._interval = p.closed(start_interval, end_interval).replace(
            left=start_inclusive, right=end_inclusive
        )

        # Raise exception if the interval duration is too short
        # Duration guaranteed to be positive thanks to the juggling above
        if self.duration <= _EPS_TIME:
            raise ValueError("Duration of the interval is negative or zero.")

    def is_in_interval(self, time):
        """
        Checks whether the requested time is within the time interval.

        Parameters
        ----------
        time : Time
            Time to be checked

        Returns
        -------
        bool
            True if time is within the reference interval, False otherwise
        """
        # check upper and lower boundaries for tolerance
        if _are_times_almost_equal(self.start, time):
            # time at starting edge - is edge closed?
            if self._interval.left:
                return True
            else:
                return False

        if _are_times_almost_equal(self.end, time):
            # time at end edge - is edge closed?
            if self._interval.right:
                return True
            else:
                return False

        # Time not edges, do a regular check
        return self._interval.contains(time)

    def is_equal(self, interval):
        """
        Checks whether two intervals are (almost) equal in value.

        If the two start or end values are as close as `_EPS_TIME`, then they are
        assumed to be equal in value.

        Parameters
        ----------
         interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        bool
            True if interval start and end are (almost) equal, False otherwise

        """
        start_equal = _are_times_almost_equal(interval.start, self.start)
        end_equal = _are_times_almost_equal(interval.end, self.end)

        if start_equal and end_equal:
            return True
        else:
            return False

    def contains(self, interval):
        """
        Checks whether the requested interval is contained within this (self) interval.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        bool
            True if check interval is contained within this interval,
            False otherwise
        """
        if self.is_in_interval(interval.start) and self.is_in_interval(interval.end):
            return True
        else:
            return False

    def is_intersecting(self, interval):
        """
        Checks whether the requested interval intersects (or is contained within)
        the reference interval.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        bool
            True if check interval intersects with the reference interval,
            False otherwise
        """
        intersection = self._interval.intersection(interval._interval)

        if intersection.empty:
            # There is absolutely no intersection
            return False

        if _are_times_almost_equal(intersection.upper, intersection.lower):
            # intersection below tolerance - practically empty intersection
            return False
        else:
            # see what the underlying function says
            return self._interval.overlaps(interval._interval)

    def intersect(self, interval):
        """
        Intersection operator for a time interval and this time interval.

        Returns a new interval that is the Intersection of two intervals,
        or None if there is no intersection.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        TimeInterval
            A new interval that is the Intersection of two intervals,
            or `None` if there is no intersection

        """
        if self.is_intersecting(interval):
            intersection = self._interval.intersection(interval._interval)

            return _create_interval_from_portion(intersection)
        else:
            return None

    def union(self, interval):
        """
        Union operator for a time interval and this time interval.

        Returns a new interval that is the Union of two intervals,
        or None if there is no intersection.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        TimeInterval
            A new interval that is the Union of two intervals,
            or `None` if there is no intersection

        """
        if self.is_intersecting(interval):
            union = self._interval.union(interval._interval)
            return _create_interval_from_portion(union)
        else:
            return None

    def expand(
        self,
        start_delta=0,
        end_delta=0,
        replicate=False,
        start_inclusive=True,
        end_inclusive=True,
    ):
        """
        Expands (or shrinks) the interval.

        Generates a new, expanded (or shrunk) `TimeInterval`, where:

        - new interval start: interval_start - start_delta
        - new interval end: interval_end + end_delta

        Negative start and/or end times are possible (to shrink the interval),
        though values ending in a negative interval will throw a `ValueError`.
        This method can be used to modify the ends of the interval (open or closed)
        as well.

        Parameters
        ----------
        start_delta : TimeDelta
            The delta time to expand the start of the interval
            (or to shrink, with negative values)
        end_delta : TimeDelta
            The delta time to expand the end of the interval
            (or to shrink, with negative values)
        replicate : bool
            `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False`
            to use a shallow copy to save memory
        start_inclusive : bool
            True if the start of the new interval is inclusive (closed), False if
            exclusive (open)
        end_inclusive : bool
            True if the start of the new interval is inclusive (closed), False if
            exclusive (open)

        Returns
        -------
        TimeInterval
            A new `TimeInterval` that is the result of the requested change

        Raises
        ------
        ValueError
            Raised if the requested expansion results in a negative duration interval
        """
        start = self.start - start_delta
        end = self.end + end_delta

        duration = end - start

        if duration > _EPS_TIME:
            return TimeInterval(
                start,
                end,
                replicate=replicate,
                start_inclusive=start_inclusive,
                end_inclusive=end_inclusive,
            )
        else:
            raise ValueError(
                "Duration of the expanded/shrunk interval is negative or zero."
            )

    @property
    def duration(self):
        """
        Computes the duration of the interval.

        Parameters
        ----------

        Returns
        -------
        duration : TimeDelta
            Duration of the interval (always positive)
        """
        return self.end - self.start

    @property
    def start(self) -> Time:
        """Returns the start time of the interval."""
        return self._interval.lower

    @property
    def end(self) -> Time:
        """Returns the end time of the interval."""
        return self._interval.upper

    def __str__(self):
        txt = ""
        if self._interval.left:
            txt += "["
        else:
            txt += "("

        txt += f" {str(getattr(self.start, self.start.format))}"
        txt += f"  {str(getattr(self.end, self.end.format))} "

        if self._interval.right:
            txt += "]"
        else:
            txt += ")"

        return txt

    @property
    def p_interval(self):
        """
        Returns the underlying `Interval` object.

        .. warning:: Most users will not need to access this object. Intended for
            developer use only.
        """
        return self._interval


def _are_times_almost_equal(t1, t2):
    """
    Checks whether two time instances are are equal to a tolerance given by `_EPS_TIME`.

    Parameters
    ----------
    t1, t2 : Time
        Times to be checked

    Returns
    -------
    True if times are almost equal, False otherwise

    """
    return abs(t1 - t2) < _EPS_TIME


def _create_interval_from_portion(interval, replicate=False):
    """
    Create a new `TimeInterval` from a given :class:`portion.interval.Interval`
    instance.

    Parameters
    ----------
    interval : Interval
        input `Interval` instance
    replicate : bool
        `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False`
        to use a shallow copy to save memory

    Returns
    -------
    TimeInterval
        New `TimeInterval` instance
    """
    return TimeInterval(
        interval.lower,
        interval.upper,
        start_inclusive=interval.left,
        end_inclusive=interval.right,
        replicate=replicate,
    )


class TimeIntervalList:
    """
    Represent and manipulate time intervals.

    This is a thin wrapper around the :class:`portion.interval.Interval` class,
    using Astropy `Time` classes under the hood.

    `start_valid` and `end_valid` values are used to mark the start and end of this
    list of time intervals. If they are not specified, the beginning and end points
    of the list of `TimeInterval` instances are used.

    Parameters
    ----------
    intervals : list[TimeInterval]
        List of intervals
    start_valid : Time or TimeInterval
        Time at which the validity of this `TimeInterval` starts (only the first
        instance contained in the `Time` object is used). If specified as TimeInterval,
        it is copied into this and `end_valid` is ignored (the only case where `None`
        is acceptable).
    end_valid : Time, TimeDelta or None
        Time or duration at which the validity of this `TimeInterval` ends (only the
        first instance contained in the `Time` object is used). None is acceptable
        only when `start_valid` is defined as a `TimeInterval`
    replicate : bool
        `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False` to
        use a shallow copy to save memory

    Raises
    ------
    ValueError
        Raised if `init_times` and `end_times` objects sizes mismatch.
    """

    def __init__(self, intervals, start_valid=None, end_valid=None, replicate=False):

        self._intervals: list = []

        if intervals:
            # if start_times is None, then there is no time interval defined

            # Fill the `Interval` list and merge as necessary
            p_intervals = self._to_p_intervals(intervals)

            # Fill the atomic `TimeInterval` objects using the merged list
            self._intervals = self._to_time_intervals(p_intervals, replicate)

        # Init range of validity
        self._valid_interval = self.__init_validity_rng(
            start_valid, end_valid, replicate
        )

    def __init_validity_rng(self, start_valid=None, end_valid=None, replicate=False):
        """
        Initialises the beginning and end of the range of validity.

        `start_valid` and `end_valid` values are used to mark the start and end of this
        list of time intervals (and are Inclusive or Closed by default).
        If they are not specified, the beginning and end points
        of the `TimeInterval` are used.

        Parameters
        ----------
        start_valid : Time or TimeInterval
            Time at which the validity of this `TimeInterval` starts (only the first
            instance contained in the `Time` object is used). If specified as
            `TimeInterval`, it is copied into this and `end_valid` is ignored
            (the only case where `None` is acceptable for it).
        end_valid : Time or TimeDelta or None
            Time or duration at which the validity of this `TimeInterval` ends (only the
            first instance contained in the `Time` object is used). None is acceptable
            only when `start_time` is defined as a `TimeInterval`
        replicate : bool
            `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False`
            to use a shallow copy to save memory

        Returns
        -------
        TimeInterval
            A new `TimeInterval` instance containing the start and end of validity
        """

        start_inclusive = True
        end_inclusive = True

        if isinstance(start_valid, TimeInterval):
            # `start_valid` is already a TimeInterval, just copy it
            return TimeInterval(start_valid, None, replicate)

        if start_valid is None:
            # No init time for validity defined, use the first interval init
            start_valid_tmp = self.get_interval(0).start
            start_inclusive = self.get_interval(0).p_interval.left
        else:
            start_valid_tmp = start_valid

        if end_valid is None:
            # No end time for validity defined, use the last interval end
            end_valid_tmp = self.get_interval(-1).end
            end_inclusive = self.get_interval(0).p_interval.right
        else:
            end_valid_tmp = end_valid

        # Generate the `TimeInterval`
        return TimeInterval(
            start_valid_tmp,
            end_valid_tmp,
            start_inclusive=start_inclusive,
            end_inclusive=end_inclusive,
            replicate=replicate,
        )

    def is_in_interval(self, time):
        """
        Checks whether the requested time is within the time interval list.

        Parameters
        ----------
        time : Time
            Time to be checked

        Returns
        -------
        bool
            `True` if time is within the interval list, `False` otherwise. Also returns
            `False` if requested time is outside validity interval.
        """
        # Is time within validity interval?
        if not self.valid_interval.is_in_interval(time):
            return False

        # Are there any events that contain this time instant?
        for interval in self._intervals:
            if interval.is_in_interval(time):
                return True

        # If we are here, then no interval contains the time
        return False

    def is_intersecting(self, interval):
        """
        Checks whether the requested interval intersects (or is contained within)
        the interval list.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        bool
            `True` if check interval intersects with the interval list,
            `False` otherwise
        """
        if len(self._intervals) == 0:
            # No interval present in the list, hence no intersection
            return False

        # While not very elegant, loop through the interval list to check
        # for intersections
        intersect_intervals = [
            interval_member
            for interval_member in self.intervals
            if interval_member.is_intersecting(interval)
        ]

        if len(intersect_intervals) > 0:
            return True
        else:
            return False

    def intersect(self, interval):
        """
        Intersection operator for a time interval and this time interval list.

        Returns a new interval that is the Intersection of the interval and
        the time interval list, or None if there is no intersection.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        TimeInterval
           A new interval that is the Intersection of the interval and
           the time interval list, or None if there is no intersection.

        """
        if len(self.intervals) == 0:
            # No interval present in the list, hence no intersection possible
            return None

        # While not very elegant, loop through the interval list to check
        # for intersections
        intersect_intervals = [
            interval_member.intersect(interval)
            for interval_member in self.intervals
            if interval_member.is_intersecting(interval)
        ]

        if len(intersect_intervals) > 0:
            # There can be only a single intersection
            return intersect_intervals[0]
        else:
            # no intersection
            return None

    def intersect_list(self, interval_list):
        """
        Intersection operator for a time interval list and this time interval list.

        Returns a new interval list that is the Intersection of `interval_list` and
        this time interval list, or None if there is no intersection even in the
        validity intervals.

        Parameters
        ----------
        interval_list : TimeIntervalList
            Time interval list to be checked

        Returns
        -------
        TimeIntervalList
           A new interval list that is the Intersection of `interval_list` and
           this time interval list, or None if there is no intersection.

        """
        if not self.valid_interval.is_intersecting(interval_list.valid_interval):
            # Validity intervals do not intersect, hence no intersection possible
            return None

        common_valid_interval = self.valid_interval.intersect(
            interval_list.valid_interval
        )

        # Compute the portion intervals
        p_self_intervals = TimeIntervalList._to_p_intervals(self.intervals)
        p_other_intervals = TimeIntervalList._to_p_intervals(interval_list.intervals)

        # Do the Intersection
        p_final = p_self_intervals.intersection(p_other_intervals)

        return TimeIntervalList(
            self._to_time_intervals(p_final), start_valid=common_valid_interval
        )

    def union(self, interval):
        """
        Union operator for a time interval and this time interval list.

        Returns a new interval that is the Union of the interval and
        the time interval list, or None if there is no intersection.

        Parameters
        ----------
        interval : TimeInterval
            Time interval to be checked

        Returns
        -------
        TimeInterval
           A new interval that is the Union of the interval and
           the time interval list, or None if there is no intersection.

        """
        if len(self.intervals) == 0:
            # No interval present in the list, hence no intersection possible
            return None

        # While not very elegant, loop through the interval list to check
        # for intersections
        union_intervals = [
            interval_member.union(interval)
            for interval_member in self.intervals
            if interval_member.is_intersecting(interval)
        ]

        if len(union_intervals) > 0:
            # There can be only a single intersection
            return union_intervals[0]
        else:
            # no intersection
            return None

    def union_list(self, interval_list):
        """
        Union operator for a time interval list and this time interval list.

        Returns a new interval list that is the Union of `interval_list` and
        this time interval list, or None if there is no intersection even in the
        validity intervals.

        Parameters
        ----------
        interval_list : TimeIntervalList
            Time interval list to be checked

        Returns
        -------
        TimeIntervalList
           A new interval list that is the Union of `interval_list` and
           this time interval list, or None if there is no intersection.

        """
        if not self.valid_interval.is_intersecting(interval_list.valid_interval):
            # Validity intervals do not intersect, hence no intersection possible
            return None

        common_valid_interval = self.valid_interval.intersect(
            interval_list.valid_interval
        )

        # Compute the portion intervals
        p_self_intervals = TimeIntervalList._to_p_intervals(self.intervals)
        p_other_intervals = TimeIntervalList._to_p_intervals(interval_list.intervals)

        # Do the Union
        p_union = p_self_intervals.union(p_other_intervals)

        # Reduce union to the common interval
        p_common = common_valid_interval.p_interval
        p_final = p_union.intersection(p_common)

        return TimeIntervalList(
            self._to_time_intervals(p_final), start_valid=common_valid_interval
        )

    def invert(self, replicate=False):
        """
        Creates an *inverted* (or complement) copy of this time interval list, while
        keeping the same validity range.

        For example, for a single interval of `[t0, t1]` in a validity interval
        `[T0,T1]`, the inverted interval list would be `[T0,t0]` and `[t1,T1]`. If
        there are no intervals, the inverse becomes the entire validity interval.

        Parameters
        ----------
        replicate : bool
            `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False`
            to use a shallow copy to save memory

        Returns
        -------
        A new `TimeIntervalList` that has the same validity range but the individual
        intervals are inverted.
        """
        # Convert `TimeInterval` list to `Interval`
        p_interval = self._to_p_intervals(self._intervals)

        # Do the inversion
        p_int_inverted = ~p_interval

        # Fix the ends as necessary - no inf allowed
        p_validity = self._to_p_intervals(self.valid_interval)
        p_int_inverted = p_int_inverted.intersection(p_validity)

        # Generate the `TimeInterval` list
        intervals = self._to_time_intervals(p_int_inverted, replicate=False)

        # Create the `TimeIntervalList` object
        return TimeIntervalList(
            intervals, start_valid=self.valid_interval, replicate=replicate
        )

    def get_interval(self, index):
        """
        Gets the time interval for the given index.

        Parameters
        ----------
        index : int
            requested index

        Returns
        -------
        TimeInterval
            `TimeInterval` corresponding to the index

        Raises
        ------
        IndexError
            Requested index is out of bounds

        """
        return self._intervals[index]

    @property
    def valid_interval(self) -> TimeInterval:
        """
        Gets the time interval of validity for the `TimeIntervalList`.
        """
        return self._valid_interval

    @property
    def intervals(self):
        """
        Gets the time intervals within this `TimeIntervalList`.
        """
        return self._intervals

    @staticmethod
    def _to_time_intervals(p_intervals, replicate=False):
        """
        Converts a `Interval` list to a `TimeInterval` list.

        Parameters
        ----------
        p_intervals : Interval
            List of intervals
        replicate : bool
            `True` to replicate (deep copy) the `Time` or `TimeDelta` objects, `False` to
            use a shallow copy to save memory

        Returns
        -------
        list[TimeInterval]
            `TimeInterval` object with the list of time intervals
        """
        intervals: list = []

        # Fill the atomic `TimeInterval` objects using the merged list
        for p_interval in p_intervals:
            intervals.append(
                _create_interval_from_portion(p_interval, replicate=replicate)
            )

        return intervals

    @staticmethod
    def _to_p_intervals(intervals):
        """
        Converts a `TimeInterval` instance or a list to an `Interval` list
        (portion library objects).

        This is usually done to merge and simplify the elements of the list.

        Parameters
        ----------
        intervals : TimeInterval or list[TimeInterval]
            List of intervals

        Returns
        -------
        `Interval` object with the list of time intervals
        """
        # Fill the `Interval` list and merge as necessary
        p_intervals = p.empty()
        if isinstance(intervals, list):
            for interval in intervals:
                p_intervals = p_intervals.union(interval.p_interval)
        else:
            # intervals object is a single TimeInterval
            p_intervals = intervals.p_interval

        return p_intervals

    def __str__(self):
        txt = ""
        if self._intervals:
            # List not empty
            for interval in self._intervals:
                txt += str(interval) + "\n"
        else:
            txt = "Time interval list is empty."

        return txt



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Base module for analytical and numerical propagators.

"""
from abc import ABC

import numpy as np
from astropy import units as u
from astropy.units import Quantity

from satmad.coordinates.trajectory import Trajectory
from satmad.core.celestial_bodies import EARTH


class AbstractPropagator(ABC):
    """
    Base class for the propagators.

    Parameters
    ----------
    name : str
        Name of the propagator
    stepsize : Quantity or TimeDelta
        output stepsize for the propagator
    central_body : CelestialBody
        Centre Celestial Body for the propagator
    """

    def __init__(self, stepsize=60 * u.s, name="Propagator", central_body=EARTH):
        self.name = name
        self._stepsize = stepsize
        self.central_body = central_body

    @property
    def stepsize(self) -> Quantity:
        """Output stepsize for the propagator."""
        return self._stepsize

    @stepsize.setter
    def stepsize(self, stepsize):
        """
        Parameters
        ----------
        stepsize : TimeDelta or Quantity
        """
        self._stepsize = stepsize

    def _generate_time_list(self, interval):
        # generate number of steps (forced to rounded up int and added one to ensure
        # time list covers the entire interval)
        no_of_steps = np.ceil((interval.duration / self.stepsize).decompose()) + 1

        # make sure there are enough elements for interpolation
        if no_of_steps < Trajectory.reqd_min_elements():
            no_of_steps = Trajectory.reqd_min_elements()

        # output time list
        time_list = interval.start + self.stepsize * np.arange(0, no_of_steps)
        time_list.format = "isot"

        return time_list



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Trajectory of an object with interpolators.

"""
import numpy as np
from astropy import units as u
from astropy.coordinates import CartesianDifferential, CartesianRepresentation, SkyCoord
from astropy.timeseries import TimeSeries

from satmad.utils.interpolators import CartInterpolator3D
from satmad.utils.timeinterval import TimeInterval


class Trajectory:
    """
    Class that keeps the coordinates of a trajectory and interpolates
    coordinates in between.

    This class keeps the coordinate points in time and lazy inits an
    interpolated a set of new
    in between coordinates when requested. This enables sub-sampling or
    supersampling the coordinates or finding special points (e.g. Equator
    crossings) with ease.

    The class uses an `InterpolatedUnivariateSpline` inside, so the spline
    function is guaranteed to pass through all provided points. The default
    spline degree is set to 5.

    """

    # Do not init interpolators by default - lazy init when requested
    _interpolators_initialised = False
    """Flag to indicate whether interpolators are already initialised."""

    _has_velocity = False
    """Flag to indicate whether there is velocity information in the supplied
    trajectory coordinates. Controls whether velocity interpolation is to
    take place or not."""

    # Default settings for the interpolator
    _spline_degree = 5
    _extrapolate_action = "raise"

    _EPS_TIME = 10 * u.us
    """Allowable time threshold, this much 'out of bounds' is allowed when
    handling requested
    interpolation times. This helps with floating point artifacts such as
    round-off errors."""

    # Shorthand to save rendering of the derived unit
    _u_km_per_s = u.km / u.s

    def __init__(self, coords_list: SkyCoord, replicate=False):
        """
        Parameters
        ----------
        coords_list : SkyCoord
            `SkyCoord` object containing the trajectory
        replicate : bool, optional
            If `True` replicates the `coords_list` into the object, otherwise
            creates a shallow copy
            of the `coords_list`. Default is False.
        """

        if replicate:
            # replicate internal coordinate data
            self._coord_list = coords_list.replicate()
        else:
            # shallow copy the internal coordinate data
            self._coord_list = coords_list

        # save the frame name for easy access when creating new instances
        self._frame_name = self._coord_list.frame[0].name

        # save the begin and end times for the internal trajectory
        # and the interpolator
        self._interval = TimeInterval(
            self._coord_list[0].obstime, self._coord_list[-1].obstime
        )

        # Check whether we have velocity data
        if self._coord_list.data.differentials:
            self._has_velocity = True

    def __call__(self, t):
        """
        Computes the interpolated coordinates at the given time(s).

        Parameters
        ----------
        t : Time
            Time or list of times where an interpolated coordinate is required

        Returns
        -------
        coords : SkyCoord
           A `SkyCoord` object with the requested coordinate(s), in the
           original frame

        Raises
        ------
        ValueError
            If the requested `t` value is out of bounds for the interpolator

        """
        if not self._interpolators_initialised:
            # Lazy init interpolators only when required
            self._init_interpolators()

        # **** get coords at requested time ****

        # compute raw r and v vectors and
        # fill Astropy cartesian vectors (with velocities if available)
        if t.isscalar:
            # there is a single time instance
            coords = CartesianRepresentation(self._compute_pos(t), copy=False)
            if self._has_velocity:
                v = CartesianDifferential(self._compute_vel(t), copy=False)
                coords = coords.with_differentials(v)
        else:
            # there is more than one time instance
            r = [self._compute_pos(time) for time in t]
            coords = CartesianRepresentation(
                np.asarray(r), unit=r[0].unit, copy=False, xyz_axis=1
            )

            if self._has_velocity:
                v = [self._compute_vel(time) for time in t]
                v = CartesianDifferential(
                    np.asarray(v), unit=v[0].unit, copy=False, xyz_axis=1
                )
                coords = coords.with_differentials(v)

        return SkyCoord(
            coords,
            obstime=t,
            frame=self._frame_name,
            representation_type="cartesian",
            differential_type="cartesian",
        )

    def _compute_pos(self, t):
        """
        Computes the position vector at the interpolated time, also checks for
        the small float artifacts at beginning or end.

        Parameters
        ----------
        t : Time
            Time or list of times where an interpolated coordinate is required

        Returns
        -------
        Position vector at the requested time.

        Raises
        ------
        ValueError
            If the requested `t` value is out of bounds for the interpolator
        """
        # check whether the requested times are _almost_ equal to begin and
        # end times.
        # this saves against floating point rounding errors
        # is the requested time very close to first time
        if abs((t - self._interval.start).to(u.us)) < self._EPS_TIME:
            # set position to first position component
            r = self._coord_list[0].cartesian.xyz

        # is the requested time very close to last time
        elif abs((t - self._interval.end).to(u.us)) < self._EPS_TIME:
            # set position to last position component
            r = self._coord_list[-1].cartesian.xyz

        # target time is either completely out of bounds or comfortably
        # inside the interpolation range
        else:
            # Convert time to "days since epoch"
            t_req = (t - self._interval.start).jd
            # interpolate to get position component at target time
            r = self._r_interpol(t_req) * u.km

        return r

    def _compute_vel(self, t):
        """
        Computes the velocity vector at the interpolated time, also checks for
        the small float artifacts at beginning or end.

        Parameters
        ----------
        t : Time
            Time or list of times where an interpolated coordinate is required

        Returns
        -------
        Velocity vector at the requested time.

        Raises
        ------
        ValueError
            If the requested `t` value is out of bounds for the interpolator
        """
        # check whether the requested times are _almost_ equal to begin and
        # end times
        # this saves against floating point rounding errors
        # is the requested time very close to first time
        if abs((t - self._interval.start).to(u.us)) < self._EPS_TIME:
            # set velocity to first velocity component
            v = self._coord_list[0].velocity.d_xyz

        # is the requested time very close to last time
        elif abs((t - self._interval.end).to(u.us)) < self._EPS_TIME:
            # set velocity to last velocity component
            v = self._coord_list[-1].velocity.d_xyz

        # target time is either completely out of bounds or comfortably
        # inside the interpolation range
        else:
            # Convert time to "days since epoch"
            t_req = (t - self._interval.start).jd

            # interpolate to get velocity  component at target time
            v = self._v_interpol(t_req) * self._u_km_per_s

        return v

    def to_timeseries(self):
        """
        Converts the internal `coord_list` of the trajectory to a table for easy
        manipulation and file output.

        Returns
        -------
        TimeSeries
            Coordinate list as `TimeSeries`
        """
        r = self.coord_list.cartesian

        # generate timeseries

        # add velocity info if possible
        if self._has_velocity:
            v = self.coord_list.velocity
            ts = TimeSeries(
                time=self.coord_list.obstime.isot,
                data={
                    "r_X": r.x,
                    "r_Y": r.y,
                    "r_Z": r.z,
                    "v_X": v.d_x,
                    "v_Y": v.d_y,
                    "v_Z": v.d_z,
                },
            )
        else:
            ts = TimeSeries(
                time=self.coord_list.obstime.isot,
                data={"r_X": r.x, "r_Y": r.y, "r_Z": r.z},
            )

        return ts

    @property
    def coord_list(self):
        """Returns the list of underlying coordinates forming the
        trajectory."""
        return self._coord_list

    @property
    def interval(self):
        """Returns the interval of validity for the
        trajectory."""
        return self._interval

    @property
    def interpolator_name(self) -> str:
        """Returns the name of the interpolator."""
        return self._r_interpol.interpolator_name

    @classmethod
    def reqd_min_elements(cls):
        """Returns the required minimum number of elements in the trajectory."""
        return cls._spline_degree

    def _init_interpolators(self):
        """
        Initialises position and velocity interpolators.
        """
        # Init time list in days
        t_list = (self._coord_list.obstime - self._interval.start).jd
        # Init pos interpolators
        self._init_pos_interpolators(t_list)

        # init velocity interpolators if velocity info present
        if self._has_velocity:
            self._init_vel_interpolators(t_list)

        # Finally, set interpolators flag
        self._interpolators_initialised = True

    def _init_pos_interpolators(self, t_list):
        """
        Initialises position interpolators (all values in km).

        Parameters
        ----------
        t_list : (N,) array_like
            Input dimension of data points -- must be strictly increasing

        """
        r_x_list = self._coord_list.cartesian.xyz[0, :].to(u.km).value
        r_y_list = self._coord_list.cartesian.xyz[1, :].to(u.km).value
        r_z_list = self._coord_list.cartesian.xyz[2, :].to(u.km).value

        self._r_interpol = CartInterpolator3D(
            t_list,
            r_x_list,
            r_y_list,
            r_z_list,
            spline_degree=self._spline_degree,
            extrapolate_action=self._extrapolate_action,
        )

    def _init_vel_interpolators(self, t_list):
        """
        Initialises velocity interpolators (all values in km/s).

        Parameters
        ----------
        t_list : (N,) array_like
            Input dimension of data points -- must be strictly increasing

        """
        v_x_list = self._coord_list.velocity.d_xyz[0].to(self._u_km_per_s).value
        v_y_list = self._coord_list.velocity.d_xyz[1].to(self._u_km_per_s).value
        v_z_list = self._coord_list.velocity.d_xyz[2].to(self._u_km_per_s).value

        self._v_interpol = CartInterpolator3D(
            t_list,
            v_x_list,
            v_y_list,
            v_z_list,
            spline_degree=self._spline_degree,
            extrapolate_action=self._extrapolate_action,
        )

    def __str__(self):
        """String representation of the object."""
        return (
            f"Trajectory from {self._interval.start} to {self._interval.end} i"
            f"n frame {self._frame_name}. "
            f"(Interpolators initialised: "
            f"{self._interpolators_initialised})"
        )



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
TLE class tests.

"""
import numpy as np
import pytest
from astropy import units as u
from astropy.time import Time
from pytest import approx

from satmad.propagation.tle import TLE


class TLEData:
    """TLE string data with name, line1 and line2"""

    def __init__(self, line1, line2, name="No Name"):
        self.line1 = line1
        self.line2 = line2
        self.name = name


@pytest.fixture
def tle_sso():
    """Test Fixture with TLE for SSO repeat groundtrack."""
    name = "SENTINEL 2A"
    line1 = "1 40697U 15028A   20164.50828565  .00000010  00000-0  20594-4 0  9999"
    line2 = "2 40697  98.5692 238.8182 0001206  86.9662 273.1664 14.30818200259759"

    return TLEData(line1, line2, name)


@pytest.fixture
def init_tle_sso(tle_sso):
    """Generates the TLE with SSO repeat groundtrack test setup."""
    return TLE.from_tle(tle_sso.line1, tle_sso.line2, tle_sso.name)


@pytest.fixture
def tle_leo():
    """Test Fixture with TLE for LEO."""
    name = "RASAT"
    line1 = "1 37791U 11044D   20160.49218133  .00000015  00000-0  11838-4 0  9998"
    line2 = "2 37791  98.1102 253.4779 0020835 300.2201  59.6937 14.64769942470937"
    return TLEData(line1, line2, name)


@pytest.fixture
def init_tle_leo(tle_leo):
    """Generates the TLE with LEO test setup."""
    return TLE.from_tle(tle_leo.line1, tle_leo.line2, tle_leo.name)


@pytest.fixture
def check_str_leo(tle_leo):
    """Generates the TLE string with LEO test setup."""
    return tle_leo.name + "\n" + tle_leo.line1 + "\n" + tle_leo.line2 + "\n"


@pytest.fixture
def tle_geo():
    """Test Fixture with GEO TLE."""
    name = "SAMPLE GEO"
    line1 = "1 99999U 12345A   20162.50918981  .00000000  00000-0  00000-0 0 00005"
    line2 = "2 99999 000.0000 124.6202 0000000 000.0000 000.0000 01.00273791000004"
    return TLEData(line1, line2, name)


@pytest.fixture
def init_tle_geo(tle_geo):
    """Generates the GEO TLE test setup."""
    return TLE.from_tle(tle_geo.line1, tle_geo.line2, tle_geo.name)


def test_tle_init_with_lines(tle_leo, check_str_leo):
    """Test initialise from TLE."""
    tle1 = TLE.from_tle(tle_leo.line1, tle_leo.line2, tle_leo.name)

    assert str(tle1) == check_str_leo


def test_tle_with_params(init_tle_leo, check_str_leo):
    """Test initialise from parameters."""

    tle1 = init_tle_leo

    # Quantity input with different units
    inclination = tle1.inclination.to(u.deg)
    # float input
    raan = tle1.raan.to_value(u.rad)
    # dimensionless units
    eccentricity = tle1.eccentricity * u.dimensionless_unscaled

    # mean motion in rad/s
    mean_motion = tle1.mean_motion._to_value(u.rad / u.s) * 1 / u.s

    # sat number given as string - should be int
    sat_num = str(tle1.sat_number)

    tle2 = TLE(
        tle1.epoch,
        inclination,
        raan,
        eccentricity,
        tle1.arg_perigee,
        tle1.mean_anomaly,
        mean_motion,
        tle1.bstar,
        tle1.n_dot,
        n_dotdot=tle1.n_dotdot,
        name=init_tle_leo.name,
        intl_designator=tle1.intl_designator,
        sat_num=sat_num,
        classification=tle1.classification,
        rev_nr=tle1.rev_nr,
        el_nr=tle1.el_nr,
    )

    assert str(tle2) == check_str_leo == str(tle1)


def test_init_geo(init_tle_geo):
    """Test init GEO satellite."""
    tle_geo = init_tle_geo

    epoch = Time("2020-06-10T12:13:14.000")
    longitude = 42.0 * u.deg

    tle = TLE.init_geo(
        epoch,
        longitude,
        name=tle_geo.name,
        sat_num=tle_geo.sat_number,
        intl_designator=tle_geo.intl_designator,
        rev_nr=tle_geo.rev_nr,
        el_nr=tle_geo.el_nr,
    )

    assert str(tle_geo) == str(tle)


def test_init_sso(init_tle_sso):
    """Test init Sun-synchronous satellite."""
    tle_sso = init_tle_sso

    epoch = tle_sso.epoch
    alt = tle_sso.sm_axis - tle_sso._earth_radius

    # Sentinel-2A has an LTAN of 22:30
    ltan = 22.5

    tle = TLE.init_sso(
        epoch,
        alt,
        ltan,
        eccentricity=tle_sso.eccentricity,
        arg_perigee=tle_sso.arg_perigee,
        mean_anomaly=tle_sso.mean_anomaly,
        earth_radius=tle_sso._earth_radius,
    )

    # print(tle_sso)
    # print(tle)

    # print(tle_sso.node_rotation_rate())
    # print(tle.node_rotation_rate())

    # print(tle_sso.raan.to(u.deg))
    # print(tle.raan.to(u.deg))

    assert tle.sm_axis.to_value(u.m) == approx(
        tle_sso.sm_axis.to_value(u.m), abs=(5e-3 * u.mm).to_value(u.m)
    )
    assert tle.inclination.to_value(u.deg) == approx(
        tle_sso.inclination.to_value(u.deg), abs=(14e-3 * u.deg).to_value()
    )

    assert tle.raan.to_value(u.deg) == approx(
        tle_sso.raan.to_value(u.deg), abs=(36e-3 * u.deg).to_value()
    )


def test_tle_init_incl_out_of_bounds(init_tle_leo):
    """Tests init with inclination input value out of bounds
    - should raise `ValueError`."""
    with pytest.raises(ValueError):
        tle1 = init_tle_leo

        TLE(
            tle1.epoch,
            2 * np.pi,
            tle1.raan,
            tle1.eccentricity,
            tle1.arg_perigee,
            tle1.mean_anomaly,
            tle1.mean_motion,
            tle1.bstar,
            tle1.n_dot,
            n_dotdot=tle1.n_dotdot,
            name=init_tle_leo.name,
            intl_designator=tle1.intl_designator,
            sat_num=tle1.sat_number,
            classification=tle1.classification,
            rev_nr=tle1.rev_nr,
            el_nr=tle1.el_nr,
        )


def test_node_rot(init_tle_sso):
    """Test orbit plane rotation rate."""
    tle = init_tle_sso

    assert tle.node_rotation_rate.to_value(u.deg / u.day) == approx(
        (0.9870658041317965 * u.deg / u.day).to_value(),
        abs=(1e-14 * u.deg / u.day).to_value(),
    )


def test_argp_rot(init_tle_sso):
    """Test argument of perigee rotation rate."""
    tle = init_tle_sso

    assert tle.argp_rotation_rate.to_value(u.deg / u.day) == approx(
        (-2.9445253809901057 * u.deg / u.day).to_value(),
        abs=(1e-14 * u.deg / u.day).to_value(),
    )


def test_incl_out_of_bounds(init_tle_leo):
    """Tests inclination setter with input value out of bounds
    - should raise `ValueError`."""
    with pytest.raises(ValueError):
        tle = init_tle_leo
        tle.inclination = 210 * u.deg


def test_raan_out_of_bounds(init_tle_leo):
    """Tests RAAN setter with input value out of bounds """
    tle = init_tle_leo
    tle.raan = 390 * u.deg
    assert tle.raan.to_value(u.deg) == approx((30 * u.deg).to_value(), abs=1e-8)


def test_getters_setters(init_tle_leo):
    """Test getters and setters of the TLE."""
    tle = init_tle_leo

    # test getters
    assert tle.sm_axis.to_value() == approx(
        (7055.953203777368 * u.km).to_value(), abs=(0.01 * u.mm).to_value(u.km)
    )
    assert tle.period.to_value(u.s) == approx(
        (5898.53720524 * u.s).to_value(), abs=(1 * u.us).to_value(u.s)
    )
    assert tle.inclination.to_value(u.deg) == approx(
        (98.1102 * u.deg).to_value(), abs=1e-6
    )
    assert tle.raan.to_value(u.deg) == approx((253.4779 * u.deg).to_value(), abs=1e-8)
    assert tle.eccentricity == approx(0.0020835, abs=1e-9)
    assert tle.mean_anomaly.to_value(u.deg) == approx(
        (59.6937 * u.deg).to_value(), abs=1e-8
    )
    assert tle.arg_perigee.to_value(u.deg) == approx(
        (300.2201 * u.deg).to_value(), abs=1e-8
    )
    assert tle.n_dot == approx(4.5451282604019e-13, rel=1e-10)

    # test setters
    tle.inclination = 30.00 * u.deg
    assert tle.inclination.to_value(u.deg) == approx(
        (30.00 * u.deg).to_value(), abs=1e-8
    )

    tle.raan = 230.00 * u.deg
    assert tle.raan.to_value(u.deg) == approx((230.00 * u.deg).to_value(), abs=1e-8)

    tle.arg_perigee = 20.00 * u.deg
    assert tle.arg_perigee.to_value(u.deg) == approx(
        (20.00 * u.deg).to_value(), abs=1e-8
    )

    tle.mean_anomaly = 20.00 * u.deg
    assert tle.mean_anomaly.to_value(u.deg) == approx(
        (20.00 * u.deg).to_value(), abs=1e-8
    )

    tle.eccentricity = 0.05
    assert tle.eccentricity == approx(0.05, abs=1e-10)

    tle.eccentricity = 0.05 * u.dimensionless_unscaled
    assert tle.eccentricity == approx(0.05, abs=1e-10)

    tle.mean_motion = 1.05e-5
    assert tle.mean_motion == 1.05e-5 * u.rad / u.s

    # setter and getter with Quantity
    tle.mean_motion = 1.05e-5 * u.rad / u.s
    assert tle.mean_motion == 1.05e-5 * u.rad / u.s

    tle.n_dot = 1.05e-5
    assert tle.n_dot == approx(1.05e-5, rel=1e-10)

    tle.el_nr = 123
    assert tle.el_nr == 123

    tle.rev_nr = 123
    assert tle.rev_nr == 123

    tle.classification = "S"
    assert tle.classification == "S"

    tle.name = "New Name"
    assert tle.name == "New Name"



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Tests the `discrete_time_intervals` package functionalities.

"""
import pytest
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation
from astropy.time import Time, TimeDelta

from satmad.coordinates.tests.test_trajectory import (
    generate_timelist,
    propagation_engine,
)
from satmad.propagation.tle import TLE
from satmad.utils.discrete_time_events import DiscreteTimeEvents


def prepare_timelist(init_time, duration, steps):
    """Prepares time list setup."""

    dt, time_list = generate_timelist(init_time, duration, steps)

    return {"dt": dt, "time_list": time_list}


def generate_leo_trajectory(time_list):
    """Generates the LEO trajectory."""

    # Init TLE
    line1 = "1 37791U 11044D   15230.18696540  .00000191  00000-0  45093-4 0  9993"
    line2 = "2 37791 098.1618 317.9053 0023092 067.3069 293.0602 14.64400558213855"

    trajectory = propagation_engine(line1, line2, time_list)["traj"]

    return trajectory


def generate_geo_42e_trajectory(time_list):
    """Generates a GEO (42 deg East) trajectory."""

    # Init TLE
    tle = TLE.init_geo(Time("2015-10-04T00:00:00.000", scale="utc"), 42 * u.deg)
    line1, line2 = tle.export_tle()

    trajectory = propagation_engine(line1, line2, time_list)["traj"]

    return trajectory


def generate_geo_140w_trajectory(time_list):
    """Generates a GEO (140 deg West) trajectory."""

    # Init TLE
    tle = TLE.init_geo(Time("2015-10-04T00:00:00.000", scale="utc"), -140 * u.deg)
    line1, line2 = tle.export_tle()

    trajectory = propagation_engine(line1, line2, time_list)["traj"]

    return trajectory


def event_engine(trajectory_generator, init_time, duration, steps):
    """Generates events list to be tested."""
    time_list = prepare_timelist(init_time, duration, steps)["time_list"]
    trajectory = trajectory_generator(time_list)

    # Init location - default Ellipsoid is WGS84
    ground_loc = EarthLocation(lat=41.015137, lon=28.979530, height=100 * u.m)

    # generate the alt-az list
    sat_alt_az_list = trajectory.coord_list.transform_to(
        AltAz(location=ground_loc, obstime=time_list)
    )

    # Find intervals in data
    events = DiscreteTimeEvents(time_list, sat_alt_az_list.alt.deg, 5)

    return events


@pytest.mark.parametrize(
    "steps, time_error, ang_error_1, ang_error_2",
    [
        (4000, 0.13 * u.ms, 9.0e-7 * u.deg, 0.02 * u.deg),
        (2000, 2.0 * u.ms, 1.6e-4 * u.deg, 0.12 * u.deg),
        (1000, 160 * u.ms, 2.1e-3 * u.deg, 3.80 * u.deg),
    ],
)
def test_basic_pass_find(steps, time_error, ang_error_1, ang_error_2):
    """Tests the basic event finding functionality."""
    # test init time
    init_time = Time("2015-10-04T00:00:00.000", scale="utc")

    duration = TimeDelta(1.0, format="jd")

    events = event_engine(generate_leo_trajectory, init_time, duration, steps)

    # print(events.start_end_intervals)
    # print(events.max_min_table)

    # Check max events
    # ----------------
    # Case 1: "2015-10-04T11:51:44.967501459", alt: 6.914915588508499 deg
    diff = (events.max_min_table[2]["value"] - 6.914915588508499) * u.deg

    # print(f"diff : {diff} --- (stepsize: {(duration/steps).to(u.s)})")
    # print(
    #     f'{events.max_min_table[2]["time"].datetime64} **** 2015-10-04T11:51:44.967501459'
    # )

    assert pytest.approx(diff.to_value(u.deg), abs=ang_error_1.to_value(u.deg)) == 0.0

    # Case 2: "2015-10-04T10:15:28.951945200", alt: 63.38630365664624 deg
    diff = (events.max_min_table[1]["value"] - 63.38630365664624) * u.deg

    # print(f"diff : {diff} --- (stepsize: {(duration/steps).to(u.s)})")
    # print(
    #     f'{events.max_min_table[1]["time"].datetime64} **** 2015-10-04T10:15:28.951945200'
    # )
    # time error:
    # - exact match for 21 sec timestep
    # - 2.5 sec off for 43 sec timestep
    # - 6.2 sec off for 86 sec timestep

    assert pytest.approx(diff.to_value(u.deg), abs=ang_error_2.to_value(u.deg)) == 0.0

    # Check start / end intervals
    # ---------------------------
    diff = (
        events.start_end_intervals.get_interval(2).start
        - Time("2015-10-04T11:49:36.742030377")
    ).to(u.ms)

    # print(f"diff : {diff} --- (stepsize: {(duration/steps).to(u.s)})")

    assert pytest.approx(diff.to_value(u.ms), abs=time_error.to_value(u.ms)) == 0.0

    diff = (
        events.start_end_intervals.get_interval(4).end
        - Time("2015-10-04T21:23:19.417927488")
    ).to(u.ms)

    # print(f"diff : {diff} --- (stepsize: {(duration/steps).to(u.s)})")

    assert pytest.approx(diff.to_value(u.ms), abs=time_error.to_value(u.ms)) == 0.0


def test_borderline():
    """Tests the borderline cases."""
    # test init time
    init_time = Time("2015-10-04T08:36:00.000", scale="utc")

    duration = Time("2015-10-04T21:20:00.000", scale="utc") - init_time

    steps = 4000

    events = event_engine(generate_leo_trajectory, init_time, duration, steps)

    print(f"search: {events.search_interval}")
    print(events.start_end_intervals)

    assert (
        str(events.start_end_intervals.get_interval(0).start)
        == "2015-10-04T08:36:00.000"
    )

    assert (
        str(events.start_end_intervals.get_interval(4).end) == "2015-10-04T21:19:48.540"
    )


def test_visibility():
    """Tests the GEO cases with continuous visibility and no visibility."""
    # test init time
    init_time = Time("2015-10-04T08:00:00.000", scale="utc")

    duration = 1 * u.day

    steps = 4000

    # GEO with full visibility
    events = event_engine(generate_geo_42e_trajectory, init_time, duration, steps)

    print(f"full vis: {events.start_end_intervals}")

    assert (
        str(events.start_end_intervals.get_interval(0).start)
        == "2015-10-04T08:00:00.000"
    )

    assert (
        str(events.start_end_intervals.get_interval(0).end) == "2015-10-05T07:59:38.400"
    )

    # GEO with no visibility
    events = event_engine(generate_geo_140w_trajectory, init_time, duration, steps)

    print(f"no vis: {events.start_end_intervals}")

    assert not events.start_end_intervals.intervals


def test_init_array_mismatch():
    """Tests the input error due to array size mismatch."""
    with pytest.raises(ValueError):
        time_list = prepare_timelist(
            Time("2015-10-04T00:00:00.000", scale="utc"), 1.0 * u.day, 1000
        )["time_list"]
        DiscreteTimeEvents(time_list, [0.1, 1.2, 2.3])



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Interpolators used throughout the project.

"""
from scipy import interpolate as ip

from satmad.utils.timeinterval import TimeInterval


class DiscreteTimeData:
    """
    Fill this
    """

    def __init__(self, time_list, value_list):
        # Check whether time_list and value_list size are equal
        if len(time_list) != len(value_list):
            raise ValueError(
                f"Time list and value list are of different sizes. "
                f"({len(time_list)} vs. ({len(value_list)}))"
            )

        # Check whether time_list and value_list size are equal
        if len(time_list) != len(value_list):
            raise ValueError(
                f"Time list and value list are of different sizes. "
                f"({len(time_list)} vs. ({len(value_list)}))"
            )

        self.data_interval = TimeInterval(time_list[0], time_list[-1])

        # Convert time to "days since epoch"
        t_float_list = (time_list - time_list[0]).jd

        # Init interpolators (for splines, root finding possible for 3rd degree
        # (cubics) only)
        self._interpolator = ip.CubicSpline(t_float_list, value_list)

        self._deriv_interpolator = self._interpolator.derivative()


class CartInterpolator3D:
    """
    One-dimensional interpolating spline for a given set of 3D data points.

    For each of the data sets, fits a spline y = spl(t) of degree
    `spline_degree` to the provided `t`, `y` data.
    Uses a `InterpolatedUnivariateSpline` inside, so the spline function
    passes through all provided points.

    Parameters
    ----------
    t : (N,) array_like
        Input dimension of data points -- must be strictly increasing
    x : (N,) array_like
        input dimension of data points
    y : (N,) array_like
        input dimension of data points
    z : (N,) array_like
        input dimension of data points
    spline_degree : int, optional
        Degree of the smoothing spline.  Must be 1 <= `k` <= 5.
    extrapolate_action : int or str, optional
        Controls the extrapolation mode for elements
        not in the interval defined by the knot sequence.

        * if ext=0 or 'extrapolate', return the extrapolated value.
        * if ext=1 or 'zeros', return 0
        * if ext=2 or 'raise', raise a ValueError
        * if ext=3 of 'const', return the boundary value.

        The default value is 0.

     See Also
     --------
     scipy.interpolate.InterpolatedUnivariateSpline : The interpolator
        used inside this class.

     Notes
     -----
     The number of data points must be larger than the `spline_degree`.
    """

    def __init__(self, t, x, y, z, spline_degree=5, extrapolate_action="raise"):
        # init interpolators
        self._r_x_interpol = ip.InterpolatedUnivariateSpline(
            t, x, k=spline_degree, ext=extrapolate_action
        )
        self._r_y_interpol = ip.InterpolatedUnivariateSpline(
            t, y, k=spline_degree, ext=extrapolate_action
        )
        self._r_z_interpol = ip.InterpolatedUnivariateSpline(
            t, z, k=spline_degree, ext=extrapolate_action
        )

        # set the interpolator class name
        self._interpolator_name = type(self._r_x_interpol).__name__

    @property
    def interpolator_name(self) -> str:
        """Returns the name of the interpolator."""
        return self._interpolator_name

    def __call__(self, t, nu=0, ext=None) -> list:
        """
         Evaluate spline (or its nu-th derivative) at positions `t`.

         Parameters
         ----------
         t : array_like
             A 1-D array of points at which to return the value of the
             smoothed spline or its derivatives. Note: t can be unordered
             but the evaluation is more efficient if t is (partially) ordered.
         nu  : int
             The order of derivative of the spline to compute.
         ext : int
             Controls the value returned for elements of ``t`` not in the
             interval defined by the knot sequence.

             * if ext=0 or 'extrapolate', return the extrapolated value.
             * if ext=1 or 'zeros', return 0
             * if ext=2 or 'raise', raise a ValueError
             * if ext=3 or 'const', return the boundary value.

             The default value is 0, passed from the initialization of
             the spline.

         Returns
         -------
         r: list
            List containing interpolated x, y and z values

        Raises
        ------
        ValueError
            If the requested `t` value is out of bounds for the interpolator
            (if `extrapolate_action` is set to `raise`)
        """

        # t limit check is carried out within the interpolators themselves

        # generate the interpolated values
        r = [
            self._r_x_interpol(t, nu, ext),
            self._r_y_interpol(t, nu, ext),
            self._r_z_interpol(t, nu, ext),
        ]

        return r

    def __str__(self):
        return (
            f"3D Cartesian Interpolator ({self.interpolator_name}) "
            f"with {len(self._r_x_interpol.get_knots())} "
            f"knots for each axis."
        )



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Satellite propagation algorithms.

"""
